from __future__ import annotations

from abc import abstractmethod
from collections import UserDict
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    Callable,
    ClassVar,
    Optional,
    Union,
)
from typing import (
    Mapping as TMapping,
)
from typing import (
    MutableMapping as TMutableMapping,
)

import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.models.utilities import (
    compute_ind_param_std_from_suff_stats,
    compute_ind_param_mean_from_suff_stats_mixture,
    compute_ind_param_std_from_suff_stats_mixture,
    compute_ind_param_std_from_suff_stats_mixture_burn_in,
    compute_probs_from_state,
)
from leaspy.utils.functional import (
    Identity,
    Mean,
    Prod,
    NamedInputFunction,
    Sqr,
    Std,
    Sum,
    SumDim,
    get_named_parameters,
)
from leaspy.utils.typing import KwargsType
from leaspy.utils.weighted_tensor import (
    TensorOrWeightedTensor,
    WeightedTensor,
    expand_left,
    sum_dim,
)

from .distributions import SymbolicDistribution
from .utilities import compute_individual_parameter_std_from_sufficient_statistics

__all__ = [
    "VariableName",
    "VariableValue",
    "VariableNameToValueMapping",
    "VariablesToFrozenSet",
    "VariablesLazyValuesRW",
    "VariablesLazyValuesRO",
    "SuffStatsRO",
    "SuffStatsRW",
    "VariableInterface",
    "IndepVariable",
    "Hyperparameter",
    "Collect",
    "ModelParameter",
    "DataVariable",
    "LatentVariableInitType",
    "LatentVariable",
    "PopulationLatentVariable",
    "IndividualLatentVariable",
    "LinkedVariable",
    "NamedVariables",
]


VariableName = str
VariableValue = TensorOrWeightedTensor[float]
VariableNameToValueMapping = TMapping[VariableName, VariableValue]
VariablesToFrozenSet = TMapping[VariableName, frozenset[VariableValue]]
VariablesLazyValuesRO = TMapping[VariableName, Optional[VariableValue]]
VariablesLazyValuesRW = TMutableMapping[VariableName, Optional[VariableValue]]
SuffStatsRO = TMapping[VariableName, torch.Tensor]  # VarValue
SuffStatsRW = TMutableMapping[VariableName, torch.Tensor]  # VarValue

LVL_IND = 0
LVL_FT = -1


class VariableInterface:
    """Interface for variable specifications."""

    is_settable: ClassVar[bool]
    """Is True if and only if state of variables is intended to be manually modified by user."""

    fixed_shape: ClassVar[bool]
    """Is True as soon as we guarantee that shape of variable is only dependent on model hyperparameters, not data."""

    @abstractmethod
    def compute(self, state: VariableNameToValueMapping) -> Optional[VariableValue]:
        """
        Compute variable value from a `state` exposing a dict-like interface: var_name -> values.

        If not relevant for variable type return None.

        Parameters
        ----------
        state : :class:`~leaspy.variables.specs.VariableNameToValueMapping`
            The state to use in order to perform computations.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableValue` :
            The variable value computed from the state.
        """

    @abstractmethod
    def get_ancestors_names(self) -> frozenset[VariableName]:
        """
        Get the names of the variables that the current variable directly depends on.

        Returns
        -------
        :obj:`frozenset` [ :class:`~leaspy.variables.specs.VariableName`] :
            The set of ancestors variable names.
        """

    # TODO? add a check or validate(value) method? (to be optionally called by State)
    # <!> should some extra context be passed to this method
    # (e.g. `n_individuals` or `n_timepoints` dimensions are not known during variable definition
    # but their consistency could/should be checked?)


class IndepVariable(VariableInterface):
    """Base class for variable that is not dependent on any other variable."""

    def get_ancestors_names(self) -> frozenset[VariableName]:
        """
        Get the names of the variables that the current variable directly depends on.

        Returns
        -------
        :obj:`frozenset` [ :class:`~leaspy.variables.specs.VariableName`] :
            The set of ancestors variable names.
        """
        return frozenset()

    def compute(self, state: VariableNameToValueMapping) -> Optional[VariableValue]:
        """
        Compute variable value from a `state` exposing a dict-like interface: var_name -> values.

        If not relevant for variable type return None.

        Parameters
        ----------
        state : :class:`~leaspy.variables.specs.VariableNameToValueMapping`
            The state to use in order to perform computations.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableValue` or None:
            The variable value computed from the state.
        """
        return None


@dataclass(frozen=True)
class Hyperparameter(IndepVariable):
    """Hyperparameters that can not be reset."""

    value: VariableValue
    """The hyperparameter value."""

    fixed_shape: ClassVar = True
    """Whether the variable has a fixed shape or not. For hyperparameters this is True."""

    is_settable: ClassVar = False
    """Whether the variable is mutable or not. For hyperparameters this is False."""

    def __post_init__(self):
        if not isinstance(self.value, (torch.Tensor, WeightedTensor)):
            object.__setattr__(self, "value", torch.tensor(self.value))

    def to_device(self, device: torch.device) -> None:
        """
        Move the value to specified device (other variables never hold values so need for this method).

        Parameters
        ----------
        device : :class:`torch.device`
            The device on which to move the variable value.
        """
        return object.__setattr__(self, "value", self.value.to(device=device))

    @property
    def shape(self) -> tuple[int, ...]:
        return self.value.shape


@dataclass(frozen=True, init=False)
class Collect:
    """
    A convenient class to produce a function to collect sufficient stats that are existing 
    or dedicated variables (to be automatically created).

    Parameters
    ----------
    existing_variables : :obj:`tuple` of :class:`~leaspy.variables.specs.VariableName`, optional
        Names of existing variables that should be included when collecting statistics.
    dedicated_variables : :obj:`dict` [:class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.LinkedVariable`], optional
        Custom or derived variables that will be included in the collection process.

    """
    existing_variables: tuple[VariableName, ...] = ()
    dedicated_variables: Optional[TMapping[VariableName, LinkedVariable]] = None

    def __init__(
        self, *existing_variables: VariableName, **dedicated_variables: LinkedVariable
    ):
        # custom init to allow more convenient variadic form
        object.__setattr__(self, "existing_variables", existing_variables)
        object.__setattr__(self, "dedicated_variables", dedicated_variables or None)

    @property
    def variables(self) -> tuple[VariableName, ...]:
        """
        Get the combined list of all variable names to be collected.

        Returns
        -------
        :obj:`tuple` of :class:`~leaspy.variables.specs.VariableName`
            Tuple containing both existing and dedicated variable names.
        """
        return self.existing_variables + tuple(self.dedicated_variables or ())

    def __call__(self, state: VariableNameToValueMapping) -> SuffStatsRW:
        """
        Collect sufficient statistics from a given state.

        Parameters
        ----------
        state : :class:`~leaspy.variables.specs.VariableNameToValueMapping`
            A mapping from variable names to their current values.

        Returns
        -------
        stats : :class:`~leaspy.variables.specs.SuffStatsRW`
            A dictionary of variable names and their corresponding values, for all variables
            defined in this collector.
        """
        return {k: state[k] for k in self.variables}


@dataclass(frozen=True)
class ModelParameter(IndepVariable):
    """
    Variable for model parameters with a maximization rule. This variable shouldn't 
    be sampled and it shouldn't be data, a hyperparameter or a linked variable.
    
    Parameters
    ----------
    shape : :obj:`tuple` of :obj:`int`
        Shape of the parameter tensor. It must be fixed and known in advance.
    suff_stats : :class:`~leaspy.variables.specs.Collect`
        A callable object that collects sufficient statistics required to compute the update.
    update_rule : :obj:`.typing.Callable` [..., :class:`~leaspy.variables.specs.VariableValue`]
        The symbolic update rule for this parameter, used during both burn-in and standard
        learning phase unless overridden by `update_rule_burn_in`. 
    update_rule_burn_in : :obj:`.typing.Callable` [..., :class:`~leaspy.variables.specs.VariableValue`] or None, optional
        An optional alternative update rule specifically used during the burn-in phase.
        If provided, it overrides `update_rule` during that phase.

    Attributes
    ----------
    _update_rule_parameters : :obj:`frozenset` of :class:`~leaspy.variables.specs.VariableName`
        Internal cache of variable names required by `update_rule`.
    _update_rule_burn_in_parameters : :obj:`frozenset` of :class:`~leaspy.variables.specs.VariableName` or None
        Internal cache of variable names required by `update_rule_burn_in`, if defined.
    fixed_shape : :obj:`bool` (class attribute)
        Indicates that this variable has a fixed shape (True by design).
    is_settable : :obj:`bool` (class attribute)
        Flags this variable as being settable externally (True by design).
    """
    shape: tuple[int, ...]
    suff_stats: Collect  # Callable[[VariablesValuesRO], SuffStatsRW]
    """
    The symbolic update functions will take variadic `suff_stats` values,
    in order to re-use NamedInputFunction logic: e.g. update_rule=Std('xi')

    <!> ISSUE: for `tau_std` and `xi_std` we also need `state` values in addition to
    `suff_stats` values (only after burn-in) since we can NOT use the variadic form
    readily for both `state` and `suff_stats` (names would be conflicting!), we sent
    `state` as a special kw variable (a bit lazy but valid) (and we prevent using this
    name for a variable as a safety)
    """

    update_rule: Callable[..., VariableValue]
    """Update rule for normal phase, and memory-less (burn-in) phase unless `update_rule_burn_in` is not None."""

    update_rule_burn_in: Optional[Callable[..., VariableValue]] = None
    """Specific rule for burn-in (currently implemented for some variables -> e.g. `xi_std`)"""

    # private attributes (computed in __post_init__)
    _update_rule_parameters: frozenset[VariableName] = field(init=False, repr=False)
    _update_rule_burn_in_parameters: Optional[frozenset[VariableName]] = field(
        default=None, init=False, repr=False
    )

    fixed_shape: ClassVar = True
    is_settable: ClassVar = True

    def __post_init__(self):
        self._check_and_store_update_rule_parameters("update_rule")
        self._check_and_store_update_rule_parameters("update_rule_burn_in")

    def _check_and_store_update_rule_parameters(self, update_method: str) -> None:
        """
        Validates and stores the keyword parameters required by the specified update rule.
        Parameters
        ----------
        update_method : :obj:`str`
            The name of the update method attribute to validate (either `"update_rule"` or
            `"update_rule_burn_in"`).

        Raises
        ------
        :exc:`LeaspyModelInputError`
            If the function associated with the `update_method` has:
            - Positional arguments
            - Unexpected keyword arguments not matching `suff_stats` variables or `'state'`
            - Any signature that cannot be parsed or is otherwise invalid
        """
        method = getattr(self, update_method)
        if method is None:
            return
        allowed_kws = set(self.suff_stats.variables).union({"state"})
        err_msg = (
            f"Function provided in `ModelParameter.{update_method}` should be a function with keyword-only parameters "
            "(using names of this variable sufficient statistics, or the special 'state' keyword): not {}"
        )
        try:
            inferred_params = get_named_parameters(method)
        except ValueError as e:
            raise LeaspyModelInputError(err_msg.format(str(e))) from e
        forbidden_kws = set(inferred_params).difference(allowed_kws)
        if len(forbidden_kws):
            raise LeaspyModelInputError(err_msg.format(forbidden_kws))

        object.__setattr__(
            self, f"_{update_method}_parameters", frozenset(inferred_params)
        )

    def compute_update(
        self,
        *,
        state: VariableNameToValueMapping,
        suff_stats: SuffStatsRO,
        burn_in: bool,
    ) -> VariableValue:
        """
        Compute the updated value for the model parameter using a maximization step.

        Parameters
        ----------
        state : :class:`~leaspy.variables.specs.VariableNameToValueMapping`
            The state to use for computations.

        suff_stats : :class:`~leaspy.variables.specs.SuffStatsRO`
            The sufficient statistics to use.

        burn_in : :obj:`bool`
            If True, use the update rule in burning phase.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableValue` :
            The computed variable value.
        """
        update_rule, update_rule_params = self.update_rule, self._update_rule_parameters
        if burn_in and self.update_rule_burn_in is not None:
            update_rule, update_rule_params = (
                self.update_rule_burn_in,
                self._update_rule_burn_in_parameters,
            )
        state_kw = dict(state=state) if "state" in update_rule_params else {}
        # <!> it would not be clean to send all suff_stats (unfiltered) for standard kw-only functions...
        return update_rule(
            **state_kw,
            **{k: suff_stats[k] for k in self._update_rule_parameters if k != "state"},
        )

    @classmethod
    def for_pop_mean(
        cls, population_variable_name: VariableName, shape: tuple[int, ...]
    ):
        """
        Smart automatic definition of `ModelParameter` when it is the mean 
        of Gaussian prior of a population latent variable.
        
        Parameters
        ----------
        population_variable_name : :class:`~leaspy.variables.specs.VariableName`
            Name of the population latent variable for which this is the prior mean.
        shape : :obj:`tuple` of :obj:`int`
            The shape of the model parameter (typically matching the variable's dimensionality).

        Returns
        -------
        :class:`~leaspy.variables.specs.ModelParameter`
            A new instance of `ModelParameter` configured as a prior mean.
        """        
        return cls(
            shape,
            suff_stats=Collect(population_variable_name),
            update_rule=Identity(population_variable_name),
        )
    
    @classmethod
    def for_ind_mean(
        cls, individual_variable_name: VariableName, shape: tuple[int, ...]
    ):
        """
        Smart automatic definition of `ModelParameter` when it is the mean 
        of Gaussian prior of an individual latent variable.
        
        Parameters
        ----------
        individual_variable_name : :class:`~leaspy.variables.specs.VariableName`
            Name of the individual latent variable for which this is the prior mean.
        shape : :obj:`tuple` of :obj:`int`
            The shape of the model parameter (typically matching the variable's dimensionality).

        Returns
        -------
        :class:`~leaspy.variables.specs.ModelParameter`
            A new instance of `ModelParameter` configured as a prior mean.
        """

        return cls(
            shape,
            suff_stats=Collect(individual_variable_name),
            update_rule=Mean(individual_variable_name, dim=LVL_IND),
        )

    @classmethod
    def for_ind_mean_mixture(cls,
                             ind_var_name: VariableName ,shape: Tuple[int, ...],):
        """
        Smart automatic definition of `ModelParameter` when it is the mean of a mixture of Gaussians
        prior of an individual latent variable.
        Extra handling to keep one mean per cluster

        Parameters
        ----------
        individual_variable_name : :class:`~leaspy.variables.specs.VariableName`
            Name of the individual latent variable for which this is the prior mean.
        shape : :obj:`tuple` of :obj:`int`
            The shape of the model parameter (typically matching the variable's dimensionality).

        Returns
        -------
        :class:`~leaspy.variables.specs.ModelParameter`
            A new instance of `ModelParameter` configured as a prior mean.
        """
        update_rule_mixture = NamedInputFunction(
            compute_ind_param_mean_from_suff_stats_mixture,
            parameters = ("state",),
            kws=dict(ip_name = ind_var_name)
        )

        return cls(
            shape,
            suff_stats=Collect(ind_var_name),
            update_rule=update_rule_mixture,
        )

   
    @classmethod
    def for_ind_std(cls, ind_var_name: VariableName, shape: Tuple[int, ...], **tol_kw):
        """
        Smart automatic definition of `ModelParameter` when it is the std-dev 
        of Gaussian prior of an individual latent variable.

        Parameters
        ----------
        ind_var_name : :class:`~leaspy.variables.specs.VariableName`
            Name of the individual latent variable for which this is the prior std-dev.
        shape : :obj:`tuple` of :obj:`int`
            The shape of the model parameter (typically matching the variable's dimensionality).

        Returns
        -------
        :class:`~leaspy.variables.specs.ModelParameter`
            A new instance of `ModelParameter` configured as a prior std-dev.
        """
        ind_var_sqr_name = f"{ind_var_name}_sqr"
        update_rule_normal = NamedInputFunction(
            compute_individual_parameter_std_from_sufficient_statistics,
            parameters=(
                "state",
                ind_var_name,
                ind_var_sqr_name,
            ),
            kws=dict(
                individual_parameter_name=ind_var_name,
                dim=LVL_IND,
                **tol_kw,
            ),
        )
        return cls(
            shape,
            suff_stats=Collect(
                ind_var_name,
                **{
                    ind_var_sqr_name: LinkedVariable(
                        Sqr(ind_var_name)
                    )
                },
            ),
            update_rule_burn_in=Std(ind_var_name, dim=LVL_IND),
            update_rule=update_rule_normal,
        )

    @classmethod
    def for_ind_std_mixture(cls, ind_var_name: VariableName, shape: Tuple[int, ...], **tol_kw):
        """
        Smart automatic definition of `ModelParameter` when it is the std-dev of Gaussian
        prior of an individual latent variable.

        Parameters
        ----------
        ind_var_name : :class:`~leaspy.variables.specs.VariableName`
            Name of the individual latent variable for which this is the prior std-dev.
        shape : :obj:`tuple` of :obj:`int`
            The shape of the model parameter (typically matching the variable's dimensionality).

        Returns
        -------
        :class:`~leaspy.variables.specs.ModelParameter`
            A new instance of `ModelParameter` configured as a prior std.
        """
        ind_var_sqr_name = f"{ind_var_name}_sqr"
        update_rule_mixture = NamedInputFunction(
            compute_ind_param_std_from_suff_stats_mixture,
            parameters=("state", ind_var_name, ind_var_sqr_name),
            kws=dict(ip_name=ind_var_name, dim=LVL_IND, **tol_kw),
        )
        update_rule_mixture_burn_in =  NamedInputFunction(
            compute_ind_param_std_from_suff_stats_mixture_burn_in,
            parameters = ("state",),
            kws=dict(ip_name = ind_var_name)
        )
        return cls(
            shape,
            suff_stats=Collect(
                ind_var_name, **{ind_var_sqr_name: LinkedVariable(Sqr(ind_var_name))}
            ),
            update_rule_burn_in=update_rule_mixture_burn_in,
            update_rule=update_rule_mixture,
        )

    @classmethod
    def for_probs(cls, shape: Tuple[int, ...],):
        """
        Smart automatic definition of `ModelParameter` when it is the probabilities of a Gaussian mixture.

        Parameters
        ----------
        shape : :obj:`tuple` of :obj:`int`
            The shape of the model parameter (typically matching the variable's dimensionality).

        Returns
        -------
        :class:`~leaspy.variables.specs.ModelParameter`
            A new instance of `ModelParameter` configured as a probability vector.
        """

        update_rule_probs = NamedInputFunction(
            compute_probs_from_state,
            parameters = ("state",),
        )

        return cls(
            shape,
            suff_stats= Collect(),
            update_rule= update_rule_probs,
        )
    
@dataclass(frozen=True)
class DataVariable(IndepVariable):
    """
    Variables for input data, that may be reset.
    
    Attributes
    ----------
    fixed_shape : :obj:`bool`
        Indicates whether the shape of the variable is fixed. For `DataVariable`.
        `False` by design, allowing for more flexible data injection.
    is_settable : :obj:`bool`
        Flag indicating whether this variable can be set/reset directly in the state.
        `True` by desdign, meaning it can be modified externally.
    """

    fixed_shape: ClassVar = False
    is_settable: ClassVar = True


class LatentVariableInitType(str, Enum):
    """
    Type of initialization for latent variables.

    Members
    -------
    PRIOR_MODE : :obj:`str`
        Initialize latent variables using the mode of their prior distribution.
    PRIOR_MEAN : :obj:`str`
        Initialize latent variables using the mean of their prior distribution.
    PRIOR_SAMPLES : :obj:`str`
        Initialize latent variables by sampling from their prior distribution.
    """

    PRIOR_MODE = "mode"
    PRIOR_MEAN = "mean"
    PRIOR_SAMPLES = "samples"


@dataclass(frozen=True)
class LatentVariable(IndepVariable):
    """
    Unobserved variable that will be sampled, with symbolic prior distribution.

    Attributes
    ----------
    prior : :class:`~leaspy.variables.distributions.SymbolicDistribution`
        The symbolic prior distribution for the latent variable (e.g. `Normal('xi_mean', 'xi_std')`).
    sampling_kws : :obj:`dict`, optional
        Optional keyword arguments to customize the sampling process (e.g. number of samples, random seed).
    
    Class Attributes
    ----------------
    is_settable : :obj:`bool`
        Indicates that this variable can be explicitly set in the model (default: True).    
    """

    # TODO/WIP? optional mask derive from optional masks of prior distribution parameters?
    # or should be fixed & explicit here?
    prior: SymbolicDistribution
    sampling_kws: Optional[KwargsType] = None

    is_settable: ClassVar = True

    def get_prior_shape(
        self, named_vars: TMapping[VariableName, VariableInterface]
    ) -> tuple[int, ...]:
        """
        Get shape of prior distribution (i.e. without any expansion for `IndividualLatentVariable`).
        
        Parameters
        ----------
        named_vars : :obj:`Mapping` [:class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.VariableInterface`]
            A mapping from variable names to their corresponding variable interfaces.
            These should include the parameters of the prior distribution.

        Returns
        -------
        :obj:`tuple` of :obj:`int`
            The shape of the prior distribution (without any replication for individual variables).

        Raises
        ------
        :exc:`LeaspyModelInputError`
            If any of the prior distribution’s parameter variables do not have a fixed shape.
        """
        bad_params = {
            n for n in self.prior.parameters_names if not named_vars[n].fixed_shape
        }
        if len(bad_params):
            raise LeaspyModelInputError(
                f"Shapes of some prior distribution parameters are not fixed: {bad_params}"
            )
        params_shapes = {n: named_vars[n].shape for n in self.prior.parameters_names}
        res = self.prior.shape(**params_shapes) # before it returned only this
        # some changes needed to handle the parameters in mixture,
        # it sampled with shape n_clusters for the individual latent parameters if we leave it as before
        # the correct sample.size is like in the classic model,
        # the latent individual variables do not have an extra dimension
        if 'Mixture' in str(self.prior):
            name = str([self.prior.parameters_names[0]])
            if 'sources' in name:
                shape_to_modif = self.prior.shape(**params_shapes)
                res = torch.Size(shape_to_modif [:1])
            if 'tau' in name:
                shape_to_modif = torch.Size([1])
                res = shape_to_modif
            if 'xi' in name:
                shape_to_modif = torch.Size([1])
                res = shape_to_modif
        return res

    def _get_init_func_generic(
        self,
        method: Union[str, LatentVariableInitType],
        *,
        sample_shape: tuple[int, ...],
    ) -> NamedInputFunction[torch.Tensor]:
        """
        Return a function that may be used for initialization.
        
        Parameters
        ----------
        method : :obj:`str` or :class:`~leaspy.variables.specs.LatentVariableInitType`
            Initialization method. Must be one of `'samples'`, `'mode'`, or `'mean'`.
        sample_shape : :obj:`tuple` of :obj:`int`
            The shape to prepend to the initialized tensor (i.e., left expansion).

        Returns
        -------
        :class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`[:class:`torch.Tensor`]
            A symbolic function to compute the initial value tensor.

        Raises
        ------
        :exc:`ValueError`
            If `method` is not one of the allowed values.
        """
        method = LatentVariableInitType(method)
        if method is LatentVariableInitType.PRIOR_SAMPLES:
            return self.prior.get_func_sample(sample_shape)
        if method is LatentVariableInitType.PRIOR_MODE:
            return self.prior.mode.then(expand_left, shape=sample_shape)
        if method is LatentVariableInitType.PRIOR_MEAN:
            return self.prior.mean.then(expand_left, shape=sample_shape)

    @abstractmethod
    def get_regularity_variables(
        self, value_name: VariableName
    ) -> dict[VariableName, LinkedVariable]:
        """Get extra linked variables to compute regularity term for this latent variable."""
        # return {
        #    # Not really useful... directly sum it to be memory efficient...
        #    f"nll_regul_{value_name}_full": LinkedVariable(
        #        self.prior.get_func_regularization(value_name)
        #    ),
        #    # TODO: jacobian as well...
        # }
        pass


class PopulationLatentVariable(LatentVariable):
    """
    Population latent variable.
    
    Attributes
    ----------
    fixed_shape : `ClassVar`[:obj:`bool`]
        Indicates that the shape is fixed (True).
    """

    # not so easy to guarantee the fixed shape property in fact...
    # (it requires that parameters of prior distribution all have fixed shapes)
    fixed_shape: ClassVar = True

    def get_init_func(
        self,
        method: Union[str, LatentVariableInitType],
    ) -> NamedInputFunction[torch.Tensor]:
        """
        Return a function that may be used for initialization.

        Parameters
        ----------
        method : :class:`~leaspy.variables.specs.LatentVariableInitType` or :obj:`str`
            The method to be used.

        Returns
        -------
        :class:`~leaspy.utils.functional._named_imput_function.NamedInputFunction`[:class:`torch.Tensor`] :
            The initialization function.
        """
        return self._get_init_func_generic(method=method, sample_shape=())

    def get_regularity_variables(
        self,
        variable_name: VariableName,
    ) -> dict[VariableName, LinkedVariable]:
        """
        Return the negative log likelihood regularity for the provided variable name.

        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable for which to retrieve regularity.

        Returns
        -------
        :obj:`dict` [ :class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.LinkedVariable`] :
            The dictionary holding the :class:`~leaspy.variables.specs.LinkedVariable` for the regularity.
        """
        # d = super().get_regularity_variables(value_name)
        d = {}
        d.update(
            {
                f"nll_regul_{variable_name}": LinkedVariable(
                    # SumDim(f"nll_regul_{value_name}_full")
                    self.prior.get_func_regularization(variable_name).then(sum_dim)
                ),
                # TODO: jacobian as well...
            }
        )
        return d


class IndividualLatentVariable(LatentVariable):
    """
    Individual latent variable.
    
    Attributes
    ----------
    fixed_shape : `ClassVar`[:obj:`bool`]
        Indicates that the shape is fixed (True).
    """

    fixed_shape: ClassVar = False

    def get_init_func(
        self,
        method: Union[str, LatentVariableInitType],
        *,
        n_individuals: int,
    ) -> NamedInputFunction[torch.Tensor]:
        """
        Return a function that may be used for initialization.

        Parameters
        ----------
        method : :class:`~leaspy.variables.specs.LatentVariableInitType` or :obj:`str`
            The method to be used.
        n_individuals : :obj:`int`
            The number of individuals, used to define the shape.

        Returns
        -------
        :class:`~leaspy.utils.functional._named_imput_function.NamedInputFunction`[:class:`torch.Tensor`] :
            The initialization function.
        """
        return self._get_init_func_generic(method=method, sample_shape=(n_individuals,))

    def get_regularity_variables(
        self,
        variable_name: VariableName,
    ) -> dict[VariableName, LinkedVariable]:
        """
        Return the negative log likelihood regularity for the provided variable name.

        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable for which to retrieve regularity.

        Returns
        -------
        :obj:`dict` [ :class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.LinkedVariable`] :
            The dictionary holding the :class:`~leaspy.variables.specs.LinkedVariable` for the regularity.
        """
        # d = super().get_regularity_variables(value_name)
        d = {}
        if 'Mixture' in str(self.prior): #specification for the mixture model : we don't want to sum all dimensions, we need one regularity per cluster
            if variable_name == 'sources' :
                d.update(
                    {
                        f"nll_regul_{variable_name}_ind": LinkedVariable(
                            self.prior.get_func_regularization(variable_name).then(
                                sum_dim, but_dim=(LVL_IND, 2) # sum per source but omit the cluster dimension as well
                            )
                        ),
                        f"nll_regul_{variable_name}": LinkedVariable(
                            SumDim(f"nll_regul_{variable_name}_ind")
                        ),
                    }
                )
            else :
                d.update(
                    {
                        f"nll_regul_{variable_name}_ind": LinkedVariable(
                            self.prior.get_func_regularization(variable_name)
                        ), # keep it per cluster dont sum all dimensions
                        f"nll_regul_{variable_name}": LinkedVariable(
                            SumDim(f"nll_regul_{variable_name}_ind")
                        ),
                    }
                )
        else:
            d.update(
                {
                    f"nll_regul_{variable_name}_ind": LinkedVariable(
                        # SumDim(f"nll_regul_{value_name}_full", but_dim=LVL_IND)
                        self.prior.get_func_regularization(variable_name).then(
                            sum_dim, but_dim=LVL_IND
                        )
                    ),
                    f"nll_regul_{variable_name}": LinkedVariable(
                        SumDim(f"nll_regul_{variable_name}_ind")
                    ),
                    # TODO: jacobian as well...
                }
            )
        return d


@dataclass(frozen=True)
class LinkedVariable(VariableInterface):
    """
    Variable which is a deterministic expression of other variables 
    (we directly use variables names instead of mappings: kws <-> vars).
    
    Parameters
    ----------
    f : :obj:`Callable`[..., :class:`~leaspy.variables.specs.VariableValue`]
        A deterministic function that computes this variable's value from its input variables.
        The function should accept keyword arguments matching the variable names in `parameters`.

    Attributes
    ----------
    parameters : :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]
        The set of variable names on which this linked variable depends.
        This is inferred internally from the function `f`.
    is_settable : `ClassVar`[:obj:`bool`]
        Indicates that this variable is not settable directly (`False`).
    fixed_shape : `ClassVar`[obj:`bool`]
        Indicates whether the shape of the linked variable is fixed.
        By design it is `False`.
    """

    f: Callable[..., VariableValue]
    parameters: frozenset[VariableName] = field(init=False)
    # expected_shape? (<!> some of the shape dimensions might not be known like `n_individuals` or `n_timepoints`...)
    # admissible_value? (<!> same issue than before, cf. remark on `IndividualLatentVariable`)

    is_settable: ClassVar = False
    # shape of linked variable may be fixed in some cases, but complex/boring/useless logic to guarantee it
    fixed_shape: ClassVar = False

    def __post_init__(self):
        try:
            inferred_params = get_named_parameters(self.f)
        except ValueError:
            raise LeaspyModelInputError(
                "Function provided in `LinkedVariable` should be a function with "
                "keyword-only parameters (using variables names)."
            )
        object.__setattr__(self, "parameters", frozenset(inferred_params))

    def get_ancestors_names(self) -> frozenset[VariableName]:
        """
        Return the set of variable names that this linked variable depends on.

        Returns
        -------
        :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]
            The names of ancestor variables used as inputs by this linked variable.
        """
        return self.parameters

    def compute(self, state: VariableNameToValueMapping) -> VariableValue:
        """
        Compute the variable value from a given State.

        Parameters
        ----------
        state : :class:`~leaspy.variables.specs.VariableNameToValueMapping`
            The state to use for computations.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableValue` :
            The value of the variable.
        """
        return self.f(**{k: state[k] for k in self.parameters})


class NamedVariables(UserDict):
    """Convenient dictionary for named variables specifications.

    In particular, it:
        1. forbids the collisions in variable names when assigning/updating the collection
        2. forbids the usage of some reserved names like 'state' or 'suff_stats'
        3. automatically adds implicit variables when variables of certain kind are added
           (e.g. dedicated vars for sufficient stats of ModelParameter)
        4. automatically adds summary variables depending on all contained variables
           (e.g. `nll_regul_ind_sum` that depends on all individual latent variables contained)

    <!> For now, you should NOT update a `NamedVariables` with another one, only update with a regular mapping.
    """

    FORBIDDEN_NAMES: ClassVar = frozenset(
        {
            "all",
            "pop",
            "ind",
            "sum",
            "tot",
            "full",
            "nll",
            "attach",
            "regul",
            "state",
            "suff_stats",
        }
    )

    AUTOMATIC_VARS: ClassVar = (
        # TODO? jacobians as well
        "nll_regul_ind_sum_ind",
        "nll_regul_ind_sum",
        # "nll_regul_pop_sum" & "nll_regul_all_sum" are not really relevant so far
        # (because priors for our population variables are NOT true bayesian priors)
        # "nll_regul_pop_sum",
        # "nll_regul_all_sum",
    )

    def __init__(self, *args, **kws):
        self._latent_pop_vars = set()
        self._latent_ind_vars = set()
        super().__init__(*args, **kws)

    def __len__(self):
        return super().__len__() + len(self.AUTOMATIC_VARS)

    def __iter__(self):
        return iter(tuple(self.data) + self.AUTOMATIC_VARS)

    def __setitem__(self, name: VariableName, var: VariableInterface) -> None:
        if name in self.FORBIDDEN_NAMES or name in self.AUTOMATIC_VARS:
            raise ValueError(f"Can not use the reserved name '{name}'")
        if name in self.data:
            raise ValueError(f"Can not reset the variable '{name}'")
        super().__setitem__(name, var)
        if isinstance(var, ModelParameter):
            self.update(var.suff_stats.dedicated_variables or {})
        if isinstance(var, LatentVariable):
            self.update(var.get_regularity_variables(name))
            if isinstance(var, PopulationLatentVariable):
                self._latent_pop_vars.add(name)
            else:
                self._latent_ind_vars.add(name)

    def __getitem__(self, name: VariableName) -> VariableInterface:
        if name in self.AUTOMATIC_VARS:
            return self._auto_vars[name]
        return super().__getitem__(name)

    @property
    def _auto_vars(self) -> dict[VariableName, LinkedVariable]:
        # TODO? add jacobian as well?
        d = dict(
            # nll_regul_pop_sum=LinkedVariable(
            #     Sum(
            #         *(
            #             f"nll_regul_{pop_var_name}"
            #             for pop_var_name in self._latent_pop_vars
            #         )
            #     )
            # ),
            nll_regul_ind_sum_ind=LinkedVariable(
                Sum(
                    *(
                        f"nll_regul_{ind_var_name}_ind"
                        for ind_var_name in self._latent_ind_vars
                    )
                )
            ),
            nll_regul_ind_sum=LinkedVariable(SumDim("nll_regul_ind_sum_ind")),
            # nll_regul_all_sum=LinkedVariable(
            #     Sum("nll_regul_pop_sum", "nll_regul_ind_sum")
            # ),
        )
        assert d.keys() == set(self.AUTOMATIC_VARS)
        return d



import warnings
from typing import Optional

from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.io.data import Dataset
from leaspy.samplers import AbstractSampler, sampler_factory
from leaspy.variables.specs import IndividualLatentVariable, PopulationLatentVariable
from leaspy.variables.state import State

from .settings import AlgorithmSettings

__all__ = ["AlgorithmWithSamplersMixin"]


class AlgorithmWithSamplersMixin:
    """Mixin class to use in algorithms that require `samplers`.

    Note that this mixin should be used with a class inheriting from `AbstractAlgo`, which must have `algo_parameters`
    attribute.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The specifications of the algorithm as a :class:`.AlgorithmSettings` instance.

        Please note that you can customize the number of memory-less (burn-in) iterations by setting either:
            * `n_burn_in_iter_frac`, such that duration of burn-in phase is a ratio of algorithm `n_iter` (default of 90%)

    Attributes
    ----------
    samplers : :obj:`dict` [:obj:`str`, :class:`~.algo.samplers.abstract_sampler.AbstractSampler` ]
        Dictionary of samplers per each variable

    current_iteration : :obj:`int`, default 0
        Current iteration of the algorithm.
        The first iteration will be 1 and the last one `n_iter`.

    random_order_variables : :obj:`bool` (default True)
        This attribute controls whether we randomize the order of variables at each iteration.
        Article https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html
        gives a rationale on why we should activate this flag.
    """

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.samplers: dict[str, AbstractSampler] = None
        self.random_order_variables: bool = self.algo_parameters.get(
            "random_order_variables", True
        )
        self.current_iteration: int = 0
        # Dynamic number of iterations for burn-in phase
        n_burn_in_iter_frac: Optional[float] = self.algo_parameters[
            "n_burn_in_iter_frac"
        ]
        if self.algo_parameters.get("n_burn_in_iter", None) is None:
            if n_burn_in_iter_frac is None:
                raise LeaspyAlgoInputError(
                    "You should NOT have both `n_burn_in_iter_frac` and `n_burn_in_iter` None."
                    "\nPlease set a value for at least one of those settings."
                )
            self.algo_parameters["n_burn_in_iter"] = int(
                n_burn_in_iter_frac * self.algo_parameters["n_iter"]
            )

        elif n_burn_in_iter_frac is not None:
            warnings.warn(
                "`n_burn_in_iter` setting is deprecated in favour of `n_burn_in_iter_frac` - "
                "which defines the duration of the burn-in phase as a ratio of the total number of iterations."
                "\nPlease use the new setting to suppress this warning or explicitly set `n_burn_in_iter_frac=None`."
                "\nNote that while `n_burn_in_iter` is supported "
                "it will always have priority over `n_burn_in_iter_frac`.",
                FutureWarning,
            )

    def _is_burn_in(self) -> bool:
        """
        Check if current iteration is in burn-in (= memory-less) phase.

        Returns
        -------
        bool
        """
        return self.current_iteration <= self.algo_parameters["n_burn_in_iter"]

    def _get_progress_str(self) -> str:
        # The algorithm must define a progress string (thanks to `self.current_iteration`)
        iter_str = super()._get_progress_str()
        if self._is_burn_in():
            iter_str += " (memory-less phase)"
        else:
            iter_str += " (with memory)"
        return iter_str

    def __str__(self):
        out = super().__str__()
        out += "\n= Samplers ="
        for sampler in self.samplers.values():
            out += f"\n    {str(sampler)}"
        return out

    def _initialize_samplers(self, state: State, dataset: Dataset) -> None:
        """
        Instantiate samplers as a dictionary samplers {variable_name: sampler}

        Parameters
        ----------
        state : :class:`.State`
        dataset : :class:`.Dataset`
        """
        self.samplers = {}
        self._initialize_population_samplers(state)
        self._initialize_individual_samplers(state, dataset.n_individuals)

    def _initialize_individual_samplers(self, state: State, n_individuals: int) -> None:
        sampler = self.algo_parameters.get("sampler_ind", None)
        if sampler is None:
            return

        # TODO: per variable and not just per type of variable?
        sampler_kws = self.algo_parameters.get("sampler_ind_params", {})
        for var_name, var in state.dag.sorted_variables_by_type[
            IndividualLatentVariable
        ].items():
            var: IndividualLatentVariable  # for type-hint only
            # remove all properties that are not currently handled by samplers and set default values
            var_kws = dict(
                var.sampling_kws or {},
                name=var_name,
                shape=var.get_prior_shape(state.dag),
            )

            # To enforce a fixed scale for a given var, one should put it in the random var specs
            # But note that for individual variables the model parameters ***_std should always be OK (> 0)
            var_kws.setdefault("scale", var.prior.stddev.call(state))

            self.samplers[var_name] = sampler_factory(
                sampler,
                IndividualLatentVariable,
                n_patients=n_individuals,
                **var_kws,
                **sampler_kws,
            )

    def _initialize_population_samplers(self, state: State) -> None:
        sampler = self.algo_parameters.get("sampler_pop", None)
        if sampler is None:
            return

        # TODO: per variable and not just per type of variable?
        sampler_kws = self.algo_parameters.get("sampler_pop_params", {})
        for var_name, var in state.dag.sorted_variables_by_type[
            PopulationLatentVariable
        ].items():
            var: PopulationLatentVariable  # for type-hint only
            # remove all properties that are not currently handled by samplers and set default values
            var_kws = dict(
                var.sampling_kws or {},
                name=var_name,
                shape=var.get_prior_shape(state.dag),
            )

            # To enforce a fixed scale for a given var, one should put it in the random var specs
            # For instance: for betas & deltas, it is a good idea to define them this way
            # since they'll probably be = 0 just after initialization!
            var_kws.setdefault("scale", state[var_name].abs())
            # TODO: after functional test passed we could change the previous line with the following one (more consistent)
            # var_kws.setdefault("scale", var.prior.stddev.call(state))

            # TODO: mask logic?

            self.samplers[var_name] = sampler_factory(
                sampler, PopulationLatentVariable, **var_kws, **sampler_kws
            )



from abc import abstractmethod
from typing import Iterable, Optional

import torch

from leaspy.utils.functional import Exp, OrthoBasis, Sqr
from leaspy.variables.distributions import Normal
from leaspy.variables.specs import (
    Hyperparameter,
    LinkedVariable,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
    SuffStatsRW,
    VariableName,
)
from leaspy.variables.state import State

from .time_reparametrized import TimeReparametrizedModel

# TODO refact? implement a single function
# compute_individual_tensorized(..., with_jacobian: bool) -> returning either
# model values or model values + jacobians wrt individual parameters

# TODO refact? subclass or other proper code technique to extract model's concrete
#  formulation depending on if linear, logistic, mixed log-lin, ...


__all__ = [
    "RiemanianManifoldModel",
    "LinearInitializationMixin",
    "LinearModel",
]


class RiemanianManifoldModel(TimeReparametrizedModel):
    """Manifold model for multiple variables of interest (logistic or linear formulation).

    Parameters
    ----------
    name : :obj:`str`
        The name of the model.
    **kwargs
        Hyperparameters of the model (including `noise_model`)

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        * If hyperparameters are inconsistent
    """

    def __init__(
        self,
        name: str,
        variables_to_track: Optional[Iterable[VariableName]] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        default_variables_to_track = [
            "g",
            "v0",
            "noise_std",
            "tau_mean",
            "tau_std",
            "xi_mean",
            "xi_std",
            "nll_attach",
            "nll_regul_log_g",
            "nll_regul_log_v0",
            "xi",
            "tau",
            "nll_regul_pop_sum",
            "nll_regul_all_sum",
            "nll_tot",
        ]
        if self.source_dimension:
            default_variables_to_track += [
                "sources",
                "betas",
                "mixing_matrix",
                "space_shifts",
            ]
        self.track_variables(variables_to_track or default_variables_to_track)

    @classmethod
    def _center_xi_realizations(cls, state: State) -> None:
        """
        Center the ``xi`` realizations in place.

        Parameters
        ----------
        state : :class:`.State`
            The dictionary-like object representing current model state, which
            contains keys such as``"xi"`` and ``"log_v0"``.

        Notes
        -----
        This transformation preserves the orthonormal basis since the new ``v0`` remains
        collinear to the previous one. It is a purely internal operation meant to reduce
        redundancy in the parameter space (i.e., improve identifiability and stabilize
        inference).
        """
        mean_xi = torch.mean(state["xi"])
        state["xi"] = state["xi"] - mean_xi
        state["log_v0"] = state["log_v0"] + mean_xi

        # TODO: find a way to prevent re-computation of orthonormal basis since it should
        #  not have changed (v0_collinear update)
        # self.update_MCMC_toolbox({'v0_collinear'}, realizations)

    @classmethod
    def compute_sufficient_statistics(cls, state: State) -> SuffStatsRW:
        """
        Compute the model's :term:`sufficient statistics`.

        Parameters
        ----------
        state : :class:`.State`
            The state to pick values from.

        Returns
        -------
        SuffStatsRW :
            The computed sufficient statistics.
        """
        # <!> modify 'xi' and 'log_v0' realizations in-place
        # TODO: what theoretical guarantees for this custom operation?
        cls._center_xi_realizations(state)

        return super().compute_sufficient_statistics(state)

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables, derived variables,
        model 'parameters') that are part of the model.

        Returns
        -------
        NamedVariables :
            A dictionary-like object mapping variable names to their specifications.
            These include `ModelParameter`, `Hyperparameter`, `PopulationLatentVariable`,
            and `LinkedVariable` instances.
        """
        d = super().get_variables_specs()
        d.update(
            # PRIORS
            log_v0_mean=ModelParameter.for_pop_mean(
                "log_v0",
                shape=(self.dimension,),
            ),
            log_v0_std=Hyperparameter(0.01),
            xi_mean=Hyperparameter(0.0),
            # LATENT VARS
            log_v0=PopulationLatentVariable(
                Normal("log_v0_mean", "log_v0_std"),
            ),
            # DERIVED VARS
            v0=LinkedVariable(
                Exp("log_v0"),
            ),
            metric=LinkedVariable(
                self.metric
            ),  # for linear model: metric & metric_sqr are fixed = 1.
        )
        if self.source_dimension >= 1:
            d.update(
                model=LinkedVariable(self.model_with_sources),
                metric_sqr=LinkedVariable(Sqr("metric")),
                orthonormal_basis=LinkedVariable(OrthoBasis("v0", "metric_sqr")),
            )
        else:
            d["model"] = LinkedVariable(self.model_no_sources)

        # TODO: WIP
        # variables_info.update(self.get_additional_ordinal_population_random_variable_information())
        # self.update_ordinal_population_random_variable_information(variables_info)

        return d

    @staticmethod
    @abstractmethod
    def metric(*, g: torch.Tensor) -> torch.Tensor:
        pass

    @classmethod
    def model_no_sources(cls, *, rt: torch.Tensor, metric, v0, g) -> torch.Tensor:
        """
        Return the model output when sources(spatial components) are not present.

        Parameters
        ----------
        rt :  :class:`torch.Tensor`
            The reparametrized time.
        metric : Any
            The metric tensor used for computing the spatial/temporal influence.
        v0 : Any
            The values of the population parameter `v0` for each feature.
        g : Any
            The values of the population parameter `g` for each feature.

        Returns
        -------
         :class:`torch.Tensor`
            The model output without contribution from source shifts.

        Notes
        -----
        This implementation delegates to `model_with_sources` with `space_shifts`
        set to a zero tensor of shape (1, 1), effectively removing source effects.
        """
        return cls.model_with_sources(
            rt=rt,
            metric=metric,
            v0=v0,
            g=g,
            space_shifts=torch.zeros((1, 1)),
        )

    @classmethod
    @abstractmethod
    def model_with_sources(
        cls,
        *,
        rt: torch.Tensor,
        space_shifts: torch.Tensor,
        metric,
        v0,
        g,
    ) -> torch.Tensor:
        pass



"""This module contains utility functions related to the functional module."""

import operator
from functools import reduce
from typing import Callable, Iterable, Optional, Set, Tuple, TypeVar

import torch

from leaspy.utils.typing import KwargsType

from ..weighted_tensor import TensorOrWeightedTensor, WeightedTensor
from ._named_input_function import NamedInputFunction

__all__ = [
    "get_named_parameters",
]

S = TypeVar("S")


def _prod(iterable: Iterable[S], start: int = 1) -> S:
    """Product of all elements of the provided iterable, starting from `start`.

    Parameters
    ----------
    iterable : :obj:`Iterable` [:class:`.S`]
        An iterable of elements to multiply.

    start : :obj:`int`, optional
        The initial value to start the product from, default is 1.

    Returns
    -------
    :class:`.S`
        The product of all elements in the iterable, starting from `start`.
    """
    return reduce(operator.mul, iterable, start)


def _prod_args(
    *args: TensorOrWeightedTensor[S], **start_kw
) -> TensorOrWeightedTensor[S]:
    """Product of tensors with variadic input instead of standard iterable input.

    Parameters
    ----------
    args : :class:`~leaspy.utils.weighted_tensor._weighted_tensors.TensorOrWeightedTensor` [:class:`.S`]
        A variable number of tensors or weighted tensors to multiply.
    start_kw : :obj:`dict`
        Additional keyword arguments to pass to the product function.

    Returns
    -------
    :obj:`~leaspy.utils.weighted_tensor._weighted_tensors.TensorOrWeightedTensor` [:class:`.S`]
        The product of all tensors in `args`, starting from the value specified in `start_kw`.
    """
    return _prod(args, **start_kw)


def _identity(x: S) -> S:
    """Unary identity function.

    Parameters
    ----------
    x : :class:`.S`
        The input value to return unchanged.

    Returns
    -------
    :class:`.S`
        The input value `x` unchanged.
    """
    return x


def get_named_parameters(f: Callable) -> Tuple[str, ...]:
    """
    Get the names of parameters of the input function `f`, which should be
    a `NamedInputFunction` or a function with keyword-only parameters.

    Parameters
    ----------
    f : :obj:`Callable`
        The function from which to extract parameter names.

    Returns
    -------
    :obj:`tuple` [:obj:`str`, ...]
        A tuple containing the names of the parameters of the function `f`.

    Raises
    ------
    :obj:`ValueError`
        If the function `f` has positional parameters or if it has keyword-only parameters
        that are not allowed by the `NamedInputFunction` interface.
    """
    from inspect import signature

    if isinstance(f, NamedInputFunction):
        return f.parameters
    params = signature(f).parameters
    non_kw_only_params = [
        p_name for p_name, p in params.items() if p.kind is not p.KEYWORD_ONLY
    ]
    # nota: we do not check annotation of returned type for now (to remain lighter)
    if len(non_kw_only_params):
        raise ValueError(non_kw_only_params)
    return tuple(params)


def _arguments_checker(
    *,
    nb_arguments: Optional[int] = None,
    mandatory_kws: Optional[Set[str]] = None,
    possible_kws: Optional[Set[str]] = None,
) -> Callable:
    """
    Factory to check basic properties of parameters names and keyword-only arguments.

    Parameters
    ----------
    nb_arguments : :obj:`int`, optional
        Fixed number of positional arguments required by the function.
    mandatory_kws : :obj:`set` :obj:`str`], optional
        Mandatory keyword-arguments for the function.
    possible_kws : :obj:`set` [:obj:`str`], optional
        Set of ALL possible keyword-arguments for the function.

    Returns
    -------
    :obj:`Callable`
        A function that checks the provided arguments and keyword arguments against the specified criteria.

    Raises
    ------
    :obj:`ValueError`
        If `nb_arguments` is not a positive integer or None, or if mandatory keyword arguments
        are not allowed, or if the number of positional arguments does not match `nb_arguments`.
    """
    if nb_arguments is not None and (
        not isinstance(nb_arguments, int) or nb_arguments < 0
    ):
        raise ValueError(
            "Number of arguments should be a positive or null integer or None. "
            f"You provided a {type(nb_arguments)}."
        )
    nb_arguments_error_msg = None
    if nb_arguments == 1:
        nb_arguments_error_msg = f"Single name expected for positional parameters"
    elif nb_arguments is not None:
        nb_arguments_error_msg = (
            f"{nb_arguments} names expected for positional parameters"
        )
    if mandatory_kws is not None and possible_kws is not None:
        unknown_mandatory_kws = mandatory_kws.difference(possible_kws)
        if len(unknown_mandatory_kws) != 0:
            raise ValueError(
                f"Some mandatory kws are not allowed: {sorted(list(unknown_mandatory_kws))}."
            )

    def check_arguments(args: tuple, kws: KwargsType) -> None:
        """Positional and keyword arguments checker.

        Parameters
        ----------
        args : :obj:`tuple`
            Positional arguments to check.
        kws : :class:`~leaspy.utils.typing.KwargsType`
            Keyword arguments to check.

        Raises
        ------
        :obj:`ValueError`
            If the number of positional arguments does not match `nb_arguments`, or if
            mandatory keyword arguments are missing, or if unknown keyword arguments are provided.

        """
        if nb_arguments_error_msg is not None:
            if len(args) != nb_arguments:
                raise ValueError(nb_arguments_error_msg)
        if mandatory_kws is not None:
            missing_kws = mandatory_kws.difference(kws)
            if len(missing_kws) != 0:
                raise ValueError(
                    f"Missing mandatory keyword-arguments: {sorted(list(missing_kws))}."
                )
        if possible_kws is not None:
            unknown_kws = set(kws).difference(possible_kws)
            if len(unknown_kws) != 0:
                raise ValueError(
                    f"Unknown keyword-arguments: {sorted(list(unknown_kws))}."
                )

    return check_arguments


def _sum_args(*args: TensorOrWeightedTensor, **start_kw) -> TensorOrWeightedTensor:
    """Summation of regular tensors with variadic input instead of standard iterable input.

    Parameters
    ----------
    args : :class:`~leaspy.utils.weighted_tensor._weighted_tensors.TensorOrWeightedTensor`
        A variable number of tensors or weighted tensors to sum.
    start_kw : :obj:`dict`
        Additional keyword arguments to pass to the sum function.

    Returns
    -------
    :obj:`~leaspy.utils.weighted_tensor._weighted_tensors.TensorOrWeightedTensor`
        The sum of all tensors in `args`, starting from the value specified in `start_kw`.
    """
    summation = sum(args, **start_kw)
    if not isinstance(summation, (torch.Tensor, WeightedTensor)):
        # If args is empty, sum returns a float 0 that needs to be converted to a tensor
        return torch.tensor(summation)
    return summation



from .base import AbstractIndividualSampler, AbstractPopulationSampler, AbstractSampler
from .factory import (
    INDIVIDUAL_SAMPLERS,
    POPULATION_SAMPLERS,
    sampler_factory,
)
from .gibbs import (
    IndividualGibbsSampler,
    PopulationFastGibbsSampler,
    PopulationGibbsSampler,
    PopulationMetropolisHastingsSampler,
)

__all__ = [
    "AbstractSampler",
    "AbstractIndividualSampler",
    "AbstractPopulationSampler",
    "IndividualGibbsSampler",
    "PopulationGibbsSampler",
    "PopulationFastGibbsSampler",
    "PopulationMetropolisHastingsSampler",
    "INDIVIDUAL_SAMPLERS",
    "POPULATION_SAMPLERS",
    "sampler_factory",
]



"""Module defining useful distribution for ordinal noise model."""

from typing import ClassVar, Optional, Union

import numpy as np
import torch
from torch.distributions.constraints import unit_interval

__all__ = [
    "discrete_sf_from_pdf",
    "compute_ordinal_pdf_from_ordinal_sf",
    "MultinomialDistribution",
    "MultinomialCdfDistribution",
]


def discrete_sf_from_pdf(pdf: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
    """
    Compute the discrete survival function values from a discrete probability density.

    For a discrete variable with levels ``l=0..L`` the survival function is
    :math:`P(X>l)=P(X\\ge l+1)` for ``l=0..L-1``. This function assumes the last
    dimension of ``pdf`` indexes the discrete levels.

    Parameters
    ----------
    pdf : :class:`torch.Tensor` or :class:`np.ndarray`
        The discrete probability density.

    Returns
    -------
    :class:`np.ndarray` :
        The discrete survival function values.
    """
    return (1 - pdf.cumsum(-1))[..., :-1]


def compute_ordinal_pdf_from_ordinal_sf(
    ordinal_sf: torch.Tensor,
    dim_ordinal_levels: int = 3,
) -> torch.Tensor:
    """
    Compute the probability density of an ordinal model from its survival function.

    Given the survival function probabilities :math:`P(X>l)=P(X\\ge l+1)` for
    ``l=0..L-1``, compute :math:`P(X=l)` for ``l=0..L``.

    Parameters
    ----------
    ordinal_sf : :class:`torch.FloatTensor`
        Survival function values : ordinal_sf[..., l] is the proba to be superior or equal to l+1
        Dimensions are:
           * 0=individual
           * 1=visit
           * 2=feature
           * 3=ordinal_level [l=0..L-1]
           * [4=individual_parameter_dim_when_gradient]
    dim_ordinal_levels : int, default = 3
        The dimension of the tensor where the ordinal levels are.

    Returns
    -------
    ordinal_pdf : :class:`torch.FloatTensor` (same shape as input, except for dimension 3 which has one more element)
        ordinal_pdf[..., l] is the proba to be equal to l (l=0..L)
    """
    # nota: torch.diff was introduced in v1.8 but would not highly improve performance of this routine anyway
    s = list(ordinal_sf.shape)
    s[dim_ordinal_levels] = 1
    last_row = torch.zeros(size=tuple(s))
    if len(s) == 5:  # in the case of gradient we added a dimension
        first_row = last_row  # gradient(P>=0) = 0
    else:
        first_row = torch.ones(size=tuple(s))  # (P>=0) = 1
    sf_sup = torch.cat([first_row, ordinal_sf], dim=dim_ordinal_levels)
    sf_inf = torch.cat([ordinal_sf, last_row], dim=dim_ordinal_levels)
    pdf = sf_sup - sf_inf
    return pdf


class MultinomialDistribution(torch.distributions.Multinomial):
    """
    Class for a multinomial distribution with only one sample based on the Multinomial torch distrib.

    Parameters
    ----------
    probs : :class:`torch.Tensor`
        The pdf of the multinomial distribution.

    Attributes
    ----------
    """

    arg_constraints: ClassVar = {}

    def __init__(self, probs, **kwargs):
        super().__init__(total_count=1, probs=probs, **kwargs)

    @classmethod
    def from_sf(cls, sf: torch.Tensor, **kws):
        """
        Generate a new MultinomialDistribution from its survival
        function instead of its probability density function.

        Parameters
        ----------
        pdf : :class:`torch.Tensor`
            The input probability density function.
        **kws
            Additional keyword arguments to be passed for instance initialization.
        """
        return cls(
            compute_ordinal_pdf_from_ordinal_sf(sf, dim_ordinal_levels=-1), **kws
        )


class MultinomialCdfDistribution(torch.distributions.Distribution):
    """
    Class for a multinomial distribution with only sample method.

    Parameters
    ----------
    sf : :class:`torch.Tensor`
        Values of the survival function [P(X > l) for l=0..L-1 where L is max_level]
        from which the distribution samples.
        Ordinal levels are assumed to be in the last dimension.
        Those values must be in [0, 1], and decreasing when ordinal level increases (not checked).
    validate_args : bool, optional (default True)
        Whether to validate the arguments or not (None for default behavior, which changed after torch >= 1.8 to True).

    Attributes
    ----------
    cdf : :class:`torch.Tensor`
        The cumulative distribution function [P(X <= l) for l=0..L] from which the distribution samples.
        The shape of latest dimension is L+1 where L is max_level.
        We always have P(X <= L) = 1
    arg_constraints : :obj:`dict`
        Contains the constraints on the arguments.
    """

    arg_constraints: ClassVar = {}

    def __init__(self, sf: torch.Tensor, *, validate_args: Optional[bool] = True):
        super().__init__(validate_args=validate_args)
        if self._validate_args:
            assert unit_interval.check(
                sf
            ).all(), "Bad probabilities in MultinomialDistribution"
        # shape of the sample (we discard the last dimension, used to store the different ordinal levels)
        self._sample_shape = sf.shape[:-1]
        # store the cumulative distribution function with trailing P(X <= L) = 1
        self.cdf = torch.cat((1.0 - sf, torch.ones((*self._sample_shape, 1))), dim=-1)

    @classmethod
    def from_pdf(cls, pdf: torch.Tensor, **kws):
        """
        Generate a new MultinomialDistribution from its probability density
        function instead of its survival function.

        Parameters
        ----------
        pdf : :class:`torch.Tensor`
            The input probability density function.
        **kws
            Additional keyword arguments to be passed for instance initialization.
        """
        return cls(discrete_sf_from_pdf(pdf), **kws)

    def sample(self) -> torch.Tensor:
        """
        Multinomial sampling.

        We sample uniformly on [0, 1( but for the latest dimension corresponding
        to ordinal levels this latest dimension will be broadcast when comparing
        with `cdf`.

        Returns
        -------
        :class:`torch.Tensor` :
            Vector of integer values corresponding to the multinomial sampling.
            Result is in [[0, L]]
        """
        r = torch.rand(self._sample_shape).unsqueeze(-1)
        return (r < self.cdf).int().argmax(dim=-1)



from ._factory import factory_weighted_tensor_unary_operator
from ._utils import (
    expand_left,
    expand_right,
    sum_dim,
    unsqueeze_right,
    wsum_dim,
    wsum_dim_return_sum_of_weights_only,
    wsum_dim_return_weighted_sum_only,
)
from ._weighted_tensor import TensorOrWeightedTensor, WeightedTensor

__all__ = [
    "expand_left",
    "expand_right",
    "factory_weighted_tensor_unary_operator",
    "sum_dim",
    "TensorOrWeightedTensor",
    "unsqueeze_right",
    "WeightedTensor",
    "wsum_dim",
    "wsum_dim_return_weighted_sum_only",
    "wsum_dim_return_sum_of_weights_only",
]



from __future__ import annotations

from typing import Callable

import torch

from leaspy.io.data.dataset import Dataset
from leaspy.models.utilities import compute_std_from_variance
from leaspy.utils.functional import Prod, Sqr
from leaspy.utils.weighted_tensor import (
    WeightedTensor,
    sum_dim,
    wsum_dim,
    wsum_dim_return_sum_of_weights_only,
    wsum_dim_return_weighted_sum_only,
)
from leaspy.variables.distributions import Normal
from leaspy.variables.specs import (
    LVL_FT,
    Collect,
    LinkedVariable,
    ModelParameter,
    VariableInterface,
    VariableName,
)
from leaspy.variables.state import State

from ._base import ObservationModel

__all__ = [
    "GaussianObservationModel",
    "FullGaussianObservationModel",
]


class GaussianObservationModel(ObservationModel):
    """
    Specialized `ObservationModel` for noisy observations with Gaussian residuals assumption.
    
    Parameters
    ----------
    name : :obj:`str`
        The name of observed variable (to name the data variable & attachment term related to this observation).
    getter : function :class:`.Dataset` -> :class:`.WeightedTensor`
        The way to retrieve the observed values from the :class:`.Dataset` (as a :class:`.WeightedTensor`):
        e.g. all values, subset of values - only x, y, z features, one-hot encoded features, ...
    loc : :obj:`str`
        The name of the variable representing the mean (location) of the Gaussian
    scale : :obj:`str`
        The name of the variable representing the standard deviation (scale) of the Gaussian (`noise_std`)
    **extra_vars : VariableInterface
        Additional variables required by the model

    Notes
    -----
    - The model uses `leaspy.variables.distributions.Normal` internally for computing
      the log-likelihood and related operations.
    """
    def __init__(
        self,
        name: VariableName,
        getter: Callable[[Dataset], WeightedTensor],
        loc: VariableName,
        scale: VariableName,
        **extra_vars: VariableInterface,
    ):
        super().__init__(name, getter, Normal(loc, scale), extra_vars=extra_vars)


class FullGaussianObservationModel(GaussianObservationModel):
    """
    Specialized `GaussianObservationModel` when all data share the same observation model, with default naming.

    The default naming is:
        - 'y' for observations
        - 'model' for model predictions
        - 'noise_std' for scale of residuals

    We also provide a convenient factory `default` for most common case, which corresponds
    to `noise_std` directly being a `ModelParameter` (it could also be a `PopulationLatentVariable`
    with positive support). Whether scale of residuals is scalar or diagonal depends on the
    `dimension` argument of this method.
    """

    tol_noise_variance = 1e-5

    def __init__(self, noise_std: VariableInterface, **extra_vars: VariableInterface):
        super().__init__(
            name="y",
            getter=self.y_getter,
            loc="model",
            scale="noise_std",
            noise_std=noise_std,
            **extra_vars,
        )

    @staticmethod
    def y_getter(dataset: Dataset) -> WeightedTensor:
        """
        Extracts the observation values and mask from a dataset.

        Parameters
        ----------
        dataset : :class:`.Dataset`
            A dataset object containing 'values' and 'mask' attributes

        Returns
        -------
        :class:`.WeightedTensor`
            A tensor containing the observed values and a boolean mask used as weights
            for likekelihood and loss computations

        Raises
        ------
        AssertionError
            If either `dataset.values`or `dataset.mask`is `None`.
        """
        assert dataset.values is not None
        assert dataset.mask is not None
        return WeightedTensor(dataset.values, weight=dataset.mask.to(torch.bool))

    @classmethod
    def noise_std_suff_stats(cls) -> dict[VariableName, LinkedVariable]:
        """
        Dictionary of sufficient statistics needed for `noise_std` (when directly a model parameter).
        
        Returns
        -------
        :obj:`dict` [ :class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.LinkedVariable`]
            A dictionary containing the sufficient statistics:
        
            - `"y_x_model"`: Product of the observed values (`"y"`) and the model predictions (`"model"`).
            - `"model_x_model"`: Squared values of the model predictions (`"model"`).
        """
        return dict(
            y_x_model=LinkedVariable(Prod("y", "model")),
            model_x_model=LinkedVariable(Sqr("model")),
        )

    @classmethod
    def scalar_noise_std_update(
        cls,
        *,
        state: State,
        y_x_model: WeightedTensor[float],
        model_x_model: WeightedTensor[float],
    ) -> torch.Tensor:
        """
        Update rule for scalar `noise_std` (when directly a model parameter), 
        from state & sufficient statistics.
        
        Computes a common `noise_std` for all the features

        Parameters
        ----------
        state: :class:`State`
            A state dictionary containing precomputed values
         y_x_model : WeightedTensor[float]
            The weighted inner product between the observations and the model predictions.
        model_x_model : WeightedTensor[float]
            The weighted inner product of the model predictions with themselves.

        Returns
        -------
       :class:`torch.Tensor`
            The updated scalar value of the `noise_std`.
        """
        y_l2 = state["y_L2"]
        n_obs = state["n_obs"]
        # TODO? by linearity couldn't we only require `-2*y_x_model + model_x_model` as summary stat?
        # and couldn't we even collect the already summed version of it?
        s1 = sum_dim(y_x_model)
        s2 = sum_dim(model_x_model)
        noise_var = (y_l2 - 2 * s1 + s2) / n_obs.float()
        return compute_std_from_variance(
            noise_var,
            varname="noise_std",
            tol=cls.tol_noise_variance,
        )

    @classmethod
    def diagonal_noise_std_update(
        cls,
        *,
        state: State,
        y_x_model: WeightedTensor[float],
        model_x_model: WeightedTensor[float],
    ) -> torch.Tensor:
        """
        Update rule for feature-wise `noise_std` (when directly a model parameter),
        from state & sufficient statistics.

        Computes one `noise_std` per feature.

        Parameters
        ----------
        state: :class:`State`
            A state dictionary containing precomputed values
        y_x_model : :class:`.WeightedTensor`[:obj:`float`]
            The weighted inner product between the observations and the model predictions.
        model_x_model : :class:`.WeightedTensor`[:obj:`float`]
            The weighted inner product of the model predictions with themselves.

        Returns
        -------
        :class:`torch.Tensor`
            The updated value of the `noise_std` for each feature.
        """
        y_l2_per_ft = state["y_L2_per_ft"]
        n_obs_per_ft = state["n_obs_per_ft"]
        # TODO: same remark as in `.scalar_noise_std_update()`
        # sum must be done after computation to use weights of y in model to mask missing data
        summed = sum_dim(-2 * y_x_model + model_x_model, but_dim=LVL_FT)
        noise_var = (y_l2_per_ft + summed) / n_obs_per_ft.float()

        return compute_std_from_variance(
            noise_var,
            varname="noise_std",
            tol=cls.tol_noise_variance,
        )

    @classmethod
    def noise_std_specs(cls, dimension: int) -> ModelParameter:
        """
        Default specifications of `noise_std` variable when directly
        modelled as a parameter (no latent population variable).

        Parameters
        ----------
        dimension : :obj:`int`
            The dimension of the `noise_std`.
            - If `dimension == 1`, a scalar `noise_std` deviation is assumed.
            - If `dimension > 1`, feature-wise independent `noise_std` deviations
            are assumed (diagonal noise).

        Returns
        -------
        ModelParameter
            The specification of the `noise_std`, including:
            - `shape`: tuple defining the parameter shape.
            - `suff_stats`: collected sufficient statistics needed for updates.
            - `update_rule`: method to update the parameter based on statistics.
        """
        update_rule = (
            cls.scalar_noise_std_update
            if dimension == 1
            else cls.diagonal_noise_std_update
        )
        return ModelParameter(
            shape=(dimension,),
            suff_stats=Collect(**cls.noise_std_suff_stats()),
            update_rule=update_rule,
        )

    @classmethod
    def with_noise_std_as_model_parameter(cls, dimension: int):
        """
        Default instance of `FullGaussianObservationModel` with `noise_std`
        (scalar or diagonal depending on `dimension`) being a `ModelParameter`.

        Parameters
        ----------
        dimension : :obj:`int`
            The dimension of the `noise_std`.
            - If `dimension == 1`, a scalar `noise_std` is assumed.
            - If `dimension > 1`, feature-wise independent `noise_std` deviations
            are assumed (diagonal noise).

        Returns
        -------
        FullGaussianObservationModel
            A configured instance with `noise_std` as a `ModelParameter`, along with the
            necessary sufficient statistics for inference.

        Raises
        ------
        ValueError
            If `dimension` is not a positive integer.
        """
        if not isinstance(dimension, int) or dimension < 1:
            raise ValueError(
                f"Dimension should be an integer >= 1. You provided {dimension}."
            )
        if dimension == 1:
            extra_vars = {
                "y_L2": LinkedVariable(
                    Sqr("y").then(wsum_dim_return_weighted_sum_only)
                ),
                "n_obs": LinkedVariable(
                    Sqr("y").then(wsum_dim_return_sum_of_weights_only)
                ),
            }
        else:
            extra_vars = {
                "y_L2_per_ft": LinkedVariable(
                    Sqr("y").then(wsum_dim_return_weighted_sum_only, but_dim=LVL_FT)
                ),
                "n_obs_per_ft": LinkedVariable(
                    Sqr("y").then(wsum_dim_return_sum_of_weights_only, but_dim=LVL_FT)
                ),
            }
        return cls(noise_std=cls.noise_std_specs(dimension), **extra_vars)

    # Util functions not directly used in code

    @classmethod
    def compute_rmse(
        cls,
        *,
        y: WeightedTensor[float],
        model: WeightedTensor[float],
    ) -> torch.Tensor:
        """
        Computes the Root Mean Square Error (RMSE) between predictions and observations.

        Parameters
        ----------
        y : :class:`.WeightedTensor`[:obj:`float`]
            The observed target values with associated weights.
        model : :class:`.WeightedTensor`[:obj:`float`]
            The model predictions with the same shape and weighting scheme as `y`.

        Returns
        -------
        :class:`torch.Tensor`
            A scalar tensor representing the RMSE between `model` and `y`.

        """
        l2: WeightedTensor[float] = (model - y) ** 2
        l2_sum, n_obs = wsum_dim(l2)
        return (l2_sum / n_obs.float()) ** 0.5

    @classmethod
    def compute_rmse_per_ft(
        cls,
        *,
        y: WeightedTensor[float],
        model: WeightedTensor[float],
    ) -> torch.Tensor:
        """
        Computes the Root Mean Square Error (RMSE) between predictions and observations
        separately for each feature.

        Parameters
        ----------
        y : :class:`.WeightedTensor`[:obj:`float`]
            The observed target values with associated weights.
        model : :class:`.WeightedTensor`[:obj:`float`]
            The model predictions with the same shape and weighting scheme as `y`.

        Returns
        -------
        :class:`torch.Tensor`
            A scalar tensor representing the RMSE between `model` and `y`.
        """
        l2: WeightedTensor[float] = (model - y) ** 2
        l2_sum_per_ft, n_obs_per_ft = wsum_dim(l2, but_dim=LVL_FT)
        return (l2_sum_per_ft / n_obs_per_ft.float()) ** 0.5

    def to_string(self) -> str:
        """method for parameter saving"""
        if self.extra_vars["noise_std"].shape == (1,):
            return "gaussian-scalar"
        return "gaussian-diagonal"



import math
import warnings
from abc import abstractmethod
from typing import Iterable, Optional, Dict

import numpy as np
import pandas as pd
import torch

from leaspy.exceptions import LeaspyInputError, LeaspyModelInputError, LeaspyIndividualParamsInputError
from leaspy.io.data.dataset import Dataset
from leaspy.models.base import InitializationMethod

from leaspy.models.obs_models import (
    FullGaussianObservationModel,
    observation_model_factory,
)
from leaspy.utils.docs import doc_with_super
from leaspy.utils.functional import Exp, MatMul, OrthoBasis, Sqr
from leaspy.utils.typing import KwargsType
from leaspy.utils.typing import DictParams, KwargsType

from leaspy.utils.weighted_tensor import (
    TensorOrWeightedTensor,
    WeightedTensor,
    unsqueeze_right,
)
from leaspy.variables.distributions import MixtureNormal, Normal
from leaspy.variables.specs import (
    Hyperparameter,
    IndividualLatentVariable,
    LinkedVariable,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
    SuffStatsRW,
    VariablesLazyValuesRO,
)

from leaspy.variables.specs import (
    LVL_FT,
)
from leaspy.variables.state import State
from .mcmc_saem_compatible import McmcSaemCompatibleModel


@doc_with_super()
class TimeReparametrizedMixtureModel(McmcSaemCompatibleModel):
    """
    A time-reparametrized model tailored to handle mixture models with multiple clusters.
    
    This class extends `TimeReparametrizedModel` to incorporate mixture-specific behaviors,
    including support for multiple clusters (`n_clusters`) and corresponding vectorized parameters.

    Parameters
    ----------
    name : :obj:`str`
        Name of the model.
    source_dimension : Optional[:obj:`int`]
        Number of sources. Dimension of spatial components (default is None).
    **kwargs: :obj:`dict`
       Additional hyperparameters for the model. Must include:
            - 'n_clusters': int
                Number of mixture components (must be ≥ 2).
            - 'dimension' or 'features': int or list
                Dimensionality of the input data.
            - 'obs_models': str, list, or dict (optional)
                Specification of the observation model(s). Defaults to "gaussian-diagonal".

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        If inconsistent hyperparameters.
    """

    _xi_mean = 0
    _xi_std = 0.5
    _tau_std = 5.0
    _noise_std = 0.1
    _sources_mean = 0
    _sources_std = 1.0

    @property
    def xi_mean(self) -> torch.Tensor:
        """Return the mean of xi as a tensor."""
        return torch.tensor([2 if i % 2 == 0 else -2 for i in range(self.n_clusters)])

    @property
    def xi_std(self) -> torch.Tensor:
        """Return the standard deviation of xi as a tensor."""
        return torch.tensor([self._xi_std] * self.n_clusters)

    @property
    def tau_std(self) -> torch.Tensor:
        """Return the standard deviation of tau as a tensor."""
        return torch.tensor([self._tau_std] * self.n_clusters)

    @property
    def noise_std(self) -> torch.Tensor:
        """Return the standard deviation of the model as a tensor."""
        return torch.tensor(self._noise_std)

    @property
    def sources_mean(self) -> torch.Tensor:
        """Return the mean of the sources as a tensor."""
        return torch.tensor([[1 if (i + j) % 2 == 0 else -1 for j in range(self.n_clusters)]
                             for i in range(self.source_dimension)])

    @property
    def sources_std(self) -> torch.Tensor:
        """Return the standard deviation of the sources as a tensor."""
        return torch.ones(
            self.source_dimension, self.n_clusters
        )

    def __init__(self, name: str, **kwargs):

        self.source_dimension: Optional[int] = None

        dimension = kwargs.get("dimension", None)
        n_clusters = kwargs.get("n_clusters", None)
        if "features" in kwargs:
            dimension = len(kwargs["features"])
        observation_models = kwargs.get("obs_models", None)
        if observation_models is None:
            observation_models = "gaussian-diagonal"
        if observation_models == "gaussian-diagonal":
            if n_clusters < 2:
                raise LeaspyInputError(
                    "Number of clusters should be at least 2 to fit a mixture model"
                )
            if dimension == 1:
                raise LeaspyInputError(
                    "You cannot use a multivariate model with 1 feature"
                )
        if isinstance(observation_models, (list, tuple)):
            kwargs["obs_models"] = tuple(
                [
                    observation_model_factory(obs_model, **kwargs)
                    for obs_model in observation_models
                ]
            )
        elif isinstance(observation_models, (dict)):
            # Not really satisfied... Used for api load
            kwargs["obs_models"] = tuple(
                [
                    observation_model_factory(
                        observation_models["y"],
                        dimension=dimension,
                        n_clusters=n_clusters,
                    )
                ]
            )
        else:
            kwargs["obs_models"] = (
                observation_model_factory(
                    observation_models, dimension=dimension, n_clusters=n_clusters
                ),
            )
        super().__init__(name, **kwargs)

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables,
        derived variables, model 'parameters') that are part of the model.

        Returns
        -------
        :class:`~leaspy.variables.specs.NamedVariables` :
            A dictionary-like object containing specifications for the variables
        """
        d = super().get_variables_specs()

        d.update(
            rt=LinkedVariable(self.time_reparametrization),
            # PRIORS
            tau_mean=ModelParameter.for_ind_mean_mixture("tau", shape=(self.n_clusters,)),
            tau_std=ModelParameter.for_ind_std_mixture("tau", shape=(self.n_clusters,)),
            xi_mean=ModelParameter.for_ind_mean_mixture("xi", shape=(self.n_clusters,)),
            xi_std=ModelParameter.for_ind_std_mixture("xi", shape=(self.n_clusters,)),
            probs = ModelParameter.for_probs(shape=self.n_clusters),
            # LATENT VARS
            xi=IndividualLatentVariable(MixtureNormal("xi_mean", "xi_std", "probs"),
                                        sampling_kws={"scale": 10},),
            tau=IndividualLatentVariable(MixtureNormal("tau_mean", "tau_std", "probs"),
                                         sampling_kws={"scale": 10},),
            # DERIVED VARS
            alpha=LinkedVariable(Exp("xi")),
        )

        if self.source_dimension >= 1:
            d.update(
                # PRIORS
                betas_mean=ModelParameter.for_pop_mean(
                    "betas",
                    shape=(self.dimension - 1, self.source_dimension),
                ),
                betas_std=Hyperparameter(0.01),
                sources_mean=ModelParameter.for_ind_mean_mixture(
                    "sources",
                    shape=(self.source_dimension, self.n_clusters,),
                ),
                sources_std=Hyperparameter(1.0),
                # LATENT VARS
                betas=PopulationLatentVariable(
                    Normal("betas_mean", "betas_std"),
                    sampling_kws={"scale": 0.5},
                ),
                sources=IndividualLatentVariable(MixtureNormal("sources_mean", "sources_std", "probs"),
                                                 sampling_kws={"scale": 10}),
                # DERIVED VARS
                mixing_matrix=LinkedVariable(
                    MatMul("orthonormal_basis", "betas").then(torch.t)
                ),  # shape: (Ns, Nfts)
                space_shifts=LinkedVariable(
                    MatMul("sources", "mixing_matrix")
                ),  # shape: (Ni, Nfts)
            )

        return d
    
    @property
    def has_sources(self) -> bool:
        """
        Indicates whether the model includes sources.

        Returns
        -------
        :obj:`bool`
            True if `source_dimension` is a positive integer.
            False otherwise.
        """
        return (
            hasattr(self, "source_dimension")
            and isinstance(self.source_dimension, int)
            and self.source_dimension > 0
        )
    
    @staticmethod
    def time_reparametrization(
        *,
        t: TensorOrWeightedTensor[float],
        alpha: torch.Tensor,
        tau: torch.Tensor,
    ) -> TensorOrWeightedTensor[float]:
        """
        Tensorized time reparametrization formula.

        .. warning::
            Shapes of tensors must be compatible between them.

        Parameters
        ----------
        t : :class:`torch.Tensor`
            Timepoints to reparametrize
        alpha : :class:`torch.Tensor`
            Acceleration factors of individual(s)
        tau : :class:`torch.Tensor`
            Time-shift(s) of individual(s)

        Returns
        -------
        :class:`torch.Tensor`
            Reparametrized time of same shape as `timepoints`
        """
        return alpha * (t - tau)


    def _validate_compatibility_of_dataset(
        self, dataset: Optional[Dataset] = None
    ) -> None:
        """
        Validate the compatibility of the provided dataset with the model's configuration.

        Parameters
        ----------
        dataset : Optional[:class:`~leaspy.io.data.Data.Dataset`], optional
            The dataset to validate against, by default None.

        Raises
        ------
        :exc: `.LeaspyModelInputError`
            If `source_dimension` is provided but not an integer in the valid range
            [0, dataset.dimension - 1), or if `n_clusters` is provided but is not an integer ≥ 2.
        """
        super()._validate_compatibility_of_dataset(dataset)

        if not dataset:
            return
        if self.source_dimension is None:
            self.source_dimension = int(dataset.dimension**0.5)
            warnings.warn(
                "You did not provide `source_dimension` hyperparameter for multivariate model, "
                f"setting it to ⌊√dimension⌋ = {self.source_dimension}."
            )
        elif not (
            isinstance(self.source_dimension, int)
            and 0 <= self.source_dimension < dataset.dimension
        ):
            raise LeaspyModelInputError(
                f"Sources dimension should be an integer in [0, dimension - 1[ "
                f"but you provided `source_dimension` = {self.source_dimension} "
                f"whereas `dimension` = {dataset.dimension}."
            )

        # add n_clusters
        if self.n_clusters is None:
            warnings.warn(
                "You did not provide `n_clusters` hyperparameter for mixture model"
            )
        elif not (isinstance(self.n_clusters, int) and self.n_clusters >= 2):
            raise LeaspyModelInputError(
                f"Number of clusters should be an integer greater than 2 "
                f"but you provided `n_clusters` = {self.n_clusters} "
            )
        
    def _audit_individual_parameters(
        self, individual_parameters: DictParams
    ) -> KwargsType:
        """
        Validate and process individual parameter inputs for model compatibility.

        Parameters
        ----------
        individual_parameters : :class:`~leaspy.utils.typing.DictParams`
            A dictionary mapping parameter names (strings) to their values,
            which can be scalars or array-like structures.

        Returns
        -------
        KwargsType: :class:`~leaspy.utils.typing.KwargsType`
            A dictionary with the following keys:
            - "nb_inds": Number of individuals
            - "tensorized_ips": Dictionary of parameters converted to 2D tensors.
            - "tensorized_ips_gen": Generator yielding tensors for each individual,
            each with an added batch dimension.

        Raises
        ------
        :exc: `LeaspyIndividualParamsInputError`
            If the provided dictionary keys do not match the expected parameter names,
            or if the sizes of individual parameters are inconsistent,
            or if `sources` parameter does not meet array-like requirements.
        """
        from .utilities import is_array_like, tensorize_2D

        expected_parameters = set(["xi", "tau"] + int(self.has_sources) * ["sources"])
        given_parameters = set(individual_parameters.keys())
        symmetric_diff = expected_parameters.symmetric_difference(given_parameters)
        if len(symmetric_diff) > 0:
            raise LeaspyIndividualParamsInputError(
                f"Individual parameters dict provided {given_parameters} "
                f"is not compatible for {self.name} model. "
                f"The expected individual parameters are {expected_parameters}."
            )
        ips_is_array_like = {
            k: is_array_like(v) for k, v in individual_parameters.items()
        }
        ips_size = {
            k: len(v) if ips_is_array_like[k] else 1
            for k, v in individual_parameters.items()
        }
        if self.has_sources:
            if not ips_is_array_like["sources"]:
                raise LeaspyIndividualParamsInputError(
                    f"Sources must be an array_like but {individual_parameters['sources']} was provided."
                )
            tau_xi_scalars = all(ips_size[k] == 1 for k in ["tau", "xi"])
            if tau_xi_scalars and (ips_size["sources"] > 1):
                # is 'sources' not a nested array? (allowed iff tau & xi are scalars)
                if not is_array_like(individual_parameters["sources"][0]):
                    # then update sources size (1D vector representing only 1 individual)
                    ips_size["sources"] = 1
            # TODO? check source dimension compatibility?
        uniq_sizes = set(ips_size.values())
        if len(uniq_sizes) != 1:
            raise LeaspyIndividualParamsInputError(
                f"Individual parameters sizes are not compatible together. Sizes are {ips_size}."
            )
        # number of individuals present
        n_individual_parameters = uniq_sizes.pop()
        # properly choose unsqueezing dimension when tensorizing array_like (useful for sources)
        # [1,2] => [[1],[2]] (expected for 2 individuals / 1D sources)
        # [1,2] => [[1,2]] (expected for 1 individual / 2D sources)
        unsqueeze_dim = 0 if n_individual_parameters == 1 else -1
        # tensorized (2D) version of ips
        tensorized_individual_parameters = {
            name: tensorize_2D(value, unsqueeze_dim=unsqueeze_dim)
            for name, value in individual_parameters.items()
        }

        return {
            "nb_inds": n_individual_parameters,
            "tensorized_ips": tensorized_individual_parameters,
            "tensorized_ips_gen": (
                {
                    name: value[individual, :].unsqueeze(0)
                    for name, value in tensorized_individual_parameters.items()
                }
                for individual in range(n_individual_parameters)
            ),
        }

    def put_individual_parameters(self, state: State, dataset: Dataset):
        """
        Initialize individual latent parameters in the given state if not already set.
        
        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            The current state object that holds all the variables
        dataset : :class:`~leaspy.io.data.Data.Dataset`
            Dataset used to initialize latent variables accordingly.
        """
        df = dataset.to_pandas().reset_index("TIME").groupby("ID").min()

        # Initialise individual parameters if they are not already initialised
        if not state.are_variables_set(("xi", "tau")):
            df_ind = df["TIME"].to_frame(name="tau")
            df_ind["xi"] = 0.0
        else:
            df_ind = pd.DataFrame(
                torch.concat([state["xi"], state["tau"]], axis=1).detach().numpy(),
                columns=["xi", "tau"],
                index=np.arange(state["xi"].shape[0]),  # use correct number of rows
            )

        if self.source_dimension > 0:
            for i in range(self.source_dimension):
                df_ind[f"sources_{i}"] = 0.0

        with state.auto_fork(None):
            state.put_individual_latent_variables(df=df_ind)

    def _load_hyperparameters(self, hyperparameters: KwargsType) -> None:
        """
        Updates all model hyperparameters from the provided hyperparameters.

        Parameters
        ----------
        hyperparameters : :class:`~leaspy.utils.typing.KwargsType`
            Dictionary containing the hyperparameters to be loaded.
            Expected keys include:
            - "features": List or sequence of feature names
            - "dimension": Integer specifying the number of features
            - "source_dimension": Integer specifying the number of sources; must be in
            [0, dimension - 1].
            - "n_clusters": Integer, must be ≥ 2

        Raises
        ------
        :exc: `LeaspyModelInputError`
            - `dimension` does not match the number of `features`
            - `source_dimension` is invalid or out of range
            - `n_clusters` is missing or less than 2
        """
        expected_hyperparameters = (
            "features",
            "dimension",
            "source_dimension",
            "n_clusters",
        )

        if "features" in hyperparameters:
            self.features = hyperparameters["features"]

        if "dimension" in hyperparameters:
            if self.features and hyperparameters["dimension"] != len(self.features):
                raise LeaspyModelInputError(
                    f"Dimension provided ({hyperparameters['dimension']}) does not match "
                    f"features ({len(self.features)})"
                )
            self.dimension = hyperparameters["dimension"]

        if "source_dimension" in hyperparameters:
            if not (
                isinstance(hyperparameters["source_dimension"], int)
                and (hyperparameters["source_dimension"] >= 0)
                and (
                    self.dimension is None
                    or hyperparameters["source_dimension"] <= self.dimension - 1
                )
            ):
                raise LeaspyModelInputError(
                    f"Source dimension should be an integer in [0, dimension - 1], "
                    f"not {hyperparameters['source_dimension']}"
                )
            self.source_dimension = hyperparameters["source_dimension"]

            if "n_clusters" in hyperparameters:
                if not (
                    isinstance(hyperparameters["n_clusters"], int)
                    and (hyperparameters["n_clusters"] >= 2)
                ):
                    raise LeaspyModelInputError(
                        f"Number of clusters should be an integer greater than 2, "
                        f"not {hyperparameters['n_clusters']} "
                    )
                self.n_clusters = hyperparameters["n_clusters"]

        self._raise_if_unknown_hyperparameters(
            expected_hyperparameters, hyperparameters
        )

    def to_dict(self, *, with_mixing_matrix: bool = True) -> KwargsType:
        """
        Export model object as dictionary ready for :term:`JSON` saving.

        Parameters
        ----------
        with_mixing_matrix : :obj:`bool` (default ``True``)
            Save the :term:`mixing matrix` in the exported file in its 'parameters' section.

            .. warning::
                It is not a real parameter and its value will be overwritten at model loading
                (orthonormal basis is recomputed from other "true" parameters and mixing matrix
                is then deduced from this orthonormal basis and the betas)!
                It was integrated historically because it is used for convenience in
                browser webtool and only there...

        Returns
        -------
        :class:`~leaspy.utils.typing.KwargsType` :
            The object as a dictionary.
        """
        # add n_clusters
        model_settings = super().to_dict()

        model_settings["n_clusters"] = self.n_clusters
        model_settings["source_dimension"] = self.source_dimension

        if with_mixing_matrix and self.source_dimension >= 1:
            # transposed compared to previous version
            model_settings["parameters"]["mixing_matrix"] = self.state[
                "mixing_matrix"
            ].tolist()

        return model_settings


@doc_with_super()
class RiemanianManifoldMixtureModel(TimeReparametrizedMixtureModel):
    """
    A riemannian manifold model tailored to handle mixture models with multiple clusters.
    
    This class extends `RiemanianManifoldModel` to incorporate mixture-specific behaviors,
    mainly the handling of sources for multiple clusters. 

    Parameters
    ----------
    name : :obj:`str`
        The name of the model.
    **kwargs
        Hyperparameters of the model (including `noise_model`)

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        * If hyperparameters are inconsistent      
    """

    def __init__(
        self, name: str, variables_to_track: Optional[Iterable[str]] = None, **kwargs
    ):
        super().__init__(name, **kwargs)

        default_variables_to_track = [
            "g",
            "v0",
            "noise_std",
            "tau_mean",
            "tau_std",
            "xi_mean",
            "xi_std",
            "nll_attach",
            "nll_regul_log_g",
            "nll_regul_log_v0",
            "xi",
            "tau",
            "nll_regul_pop_sum",
            "nll_regul_all_sum",
            "nll_tot",
        ]

        if self.source_dimension:
            default_variables_to_track += [
                "sources",
                "betas",
                "mixing_matrix",
                "space_shifts",
                "sources_mean",
            ]

        variables_to_track = variables_to_track or default_variables_to_track
        self.tracked_variables = self.tracked_variables.union(set(variables_to_track))

        self.tracked_variables_ordered = variables_to_track

    @classmethod
    def _center_xi_realizations(cls, state: State) -> None:
        """
        Center the ``xi`` realizations in place.

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            The dictionary-like object representing current model state, which
            contains keys such as``"xi"`` and ``"log_v0"``.

        Notes
        -----
        This transformation preserves the orthonormal basis since the new ``v0`` remains
        collinear to the previous one. It is a purely internal operation meant to reduce
        redundancy in the parameter space (i.e., improve identifiability and stabilize
        inference).
        """
        mean_xi = torch.mean(state["xi"])
        state["xi"] = state["xi"] - mean_xi
        state["log_v0"] = state["log_v0"] + mean_xi

    @classmethod
    def _center_sources_realizations(cls, state: State) -> None:
        """
        Center the ``sources`` realizations in place.

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            The dictionary-like object representing current model state, which
            contains keys such as``"sources"``.
        """
        mean_sources = torch.mean(state["sources"])
        state["sources"] = state["sources"] - mean_sources

    @classmethod
    def compute_sufficient_statistics(cls, state: State) -> SuffStatsRW:
        """
        Compute the model's :term:`sufficient statistics`.

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            The state to pick values from.

        Returns
        -------
        SuffStatsRW :
            The computed sufficient statistics.
        """
        cls._center_xi_realizations(state)
        cls._center_sources_realizations(state)

        return super().compute_sufficient_statistics(state)

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables, derived variables,
        model 'parameters') that are part of the model.

        Returns
        -------
        :class:`~leaspy.variables.specs.NamedVariables`
            A dictionary-like object mapping variable names to their specifications.
            These include `ModelParameter`, `Hyperparameter`, `PopulationLatentVariable`,
            and `LinkedVariable` instances.
        """
        d = super().get_variables_specs()
        d.update(
            # PRIORS
            log_v0_mean=ModelParameter.for_pop_mean(
                "log_v0",
                shape=(self.dimension,),
            ),
            log_v0_std=Hyperparameter(0.01),
            # no xi_mean as hyperaparameter
            # LATENT VARS
            log_v0=PopulationLatentVariable(Normal("log_v0_mean", "log_v0_std")),
            # DERIVED VARS
            v0=LinkedVariable(
                Exp("log_v0"),
            ),
            metric=LinkedVariable(
                self.metric
            ),  # for linear model: metric & metric_sqr are fixed = 1.
        )

        if self.source_dimension >= 1:
            d.update(
                model=LinkedVariable(self.model_with_sources),
                metric_sqr=LinkedVariable(Sqr("metric")),
                orthonormal_basis=LinkedVariable(OrthoBasis("v0", "metric_sqr")),
            )
        else:
            d["model"] = LinkedVariable(self.model_no_sources)

        return d

    @staticmethod
    @abstractmethod
    def metric(*, g: torch.Tensor) -> torch.Tensor:
        pass

    @classmethod
    def model_no_sources(cls, *, rt: torch.Tensor, metric, v0, g) -> torch.Tensor:
        """
        Return the model output when sources(spatial components) are not present.

        Parameters
        ----------
        rt :  :class:`torch.Tensor`
            The reparametrized time.
        metric : Any
            The metric tensor used for computing the spatial/temporal influence.
        v0 : Any
            The values of the population parameter `v0` for each feature.
        g : Any
            The values of the population parameter `g` for each feature.

        Returns
        -------
         :class:`torch.Tensor`
            The model output without contribution from source shifts.

        Notes
        -----
        This implementation delegates to `model_with_sources` with `space_shifts`
        set to a zero tensor of shape (1, 1), effectively removing source effects.
        """
        return cls.model_with_sources(
            rt=rt,
            metric=metric,
            v0=v0,
            g=g,
            space_shifts=torch.zeros((1, 1)),
        )

    @classmethod
    @abstractmethod
    def model_with_sources(
        cls,
        *,
        rt: torch.Tensor,
        space_shifts: torch.Tensor,
        metric,
        v0,
        g,
    ) -> torch.Tensor:
        pass


class LogisticMixtureInitializationMixin:
    def _compute_initial_values_for_model_parameters(
        self,
        dataset: Dataset,
        ) -> VariablesLazyValuesRO:
        """
        Compute initial values for model parameters.

        Parameters
        ----------
        dataset : ::class:`~leaspy.io.data.Data.Dataset`
            The dataset from which to extract observations and masks.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariablesLazyValuesRO`
            A dictionary mapping parameter names (as strings) to their initialized
            torch.Tensor values.

        Notes
        -----
        - If the initialization method is `DEFAULT`, patient means are used.
        - If `RANDOM`, parameters are sampled from normal distributions
        centered at patient means with estimated standard deviations.
        - `values` are clamped between 0.01 and 0.99 to avoid boundary issues.
        - If the model includes sources (source_dimension >= 1),
        regression coefficients `betas_mean` are initialized accordingly.
        - If the observation model is a `FullGaussianObservationModel`,
        the noise standard deviation parameter is expanded to the correct shape.
        """
        from leaspy.models.utilities import (
            compute_patient_slopes_distribution,
            compute_patient_time_distribution,
            compute_patient_values_distribution,
            get_log_velocities,
            torch_round,
        )

        # initialize a df with the probabilities of each individual belonging to each cluster
        df = dataset.to_pandas(apply_headers=True)
        n_inds = df.reset_index("TIME").groupby("ID").min().shape[0]
        n_clusters = self.n_clusters
        probs = torch.ones(n_clusters) / n_clusters

        slopes_mu, slopes_sigma = compute_patient_slopes_distribution(df)
        values_mu, values_sigma = compute_patient_values_distribution(df)

        if self.initialization_method == InitializationMethod.DEFAULT:
            slopes = slopes_mu
            values = values_mu
            betas = torch.zeros((self.dimension - 1, self.source_dimension))

        if self.initialization_method == InitializationMethod.RANDOM:
            slopes = torch.normal(slopes_mu, slopes_sigma)
            values = torch.normal(values_mu, values_sigma)
            betas = torch.distributions.normal.Normal(loc=0.0, scale=1.0).sample(
                sample_shape=(self.dimension - 1, self.source_dimension)
            )

        step = math.ceil(n_inds / n_clusters)
        start = 0
        ids = pd.DataFrame(
            df.index.get_level_values("ID").unique()
        )  # get the values of the IDs

        for c in range(n_clusters):
            ids_cluster = ids.loc[
                start : step * (c + 1), "ID"
            ]  # get the IDs of the cluster
            df_cluster = df.loc[
                ids_cluster.values
            ]  # get all the dataframe for the cluster
            time_mu, time_sigma = compute_patient_time_distribution(df_cluster)

            if self.initialization_method == InitializationMethod.DEFAULT:
                t0_c = time_mu

            if self.initialization_method == InitializationMethod.RANDOM:
                t0_c = torch.normal(time_mu, time_sigma)

            start = step * (c + 1) + 1

            # stock the values for all the clusters
            if c == 0:
                t0 = t0_c.unsqueeze(-2)
            else:
                t0 = torch.tensor(np.append(t0, t0_c.item()))

        # Enforce values are between 0 and 1
        values = values.clamp(
            min=1e-2, max=1 - 1e-2
        )  # always "works" for ordinal (values >= 1)

        parameters = {
            "log_g_mean": torch.log(1.0 / values - 1.0),
            "log_v0_mean": get_log_velocities(slopes, self.features),
            "tau_mean": t0,
            "tau_std": self.tau_std,
            "xi_mean": self.xi_mean,
            "xi_std": self.xi_std,
            "probs": probs,
        }
        if self.source_dimension >= 1:
            parameters["betas_mean"] = betas
            parameters["sources_mean"] = self.sources_mean
            rounded_parameters = {
                str(p): torch_round(v.to(torch.float32)) for p, v in parameters.items()
            }
            obs_model = next(iter(self.obs_models))  # WIP: multiple obs models...
            if isinstance(obs_model, FullGaussianObservationModel):
                rounded_parameters["noise_std"] = self.noise_std.expand(
                    obs_model.extra_vars["noise_std"].shape
                )
            return rounded_parameters


class LogisticMultivariateMixtureModel(
    LogisticMixtureInitializationMixin, RiemanianManifoldMixtureModel
):
    """Mixture Manifold model for multiple variables of interest (logistic formulation)."""

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables, derived variables,
        model 'parameters') that are part of the model.

        Returns
        -------
        :class:`~leaspy.variables.specs.NamedVariables`
            A dictionary-like object mapping variable names to their specifications.
            These include `ModelParameter`, `Hyperparameter`, `PopulationLatentVariable`,
            and `LinkedVariable` instances.
        """
        d = super().get_variables_specs()
        d.update(
            log_g_mean=ModelParameter.for_pop_mean("log_g", shape=(self.dimension,)),
            log_g_std=Hyperparameter(0.01),
            log_g=PopulationLatentVariable(Normal("log_g_mean", "log_g_std")),
            g=LinkedVariable(Exp("log_g")),
        )
        return d

    @staticmethod
    def metric(*, g: torch.Tensor) -> torch.Tensor:
        """
        Compute the metric tensor from input tensor `g`.
        This function calculates the metric as \((g + 1)^2 / g\) element-wise.

        Parameters
        ----------
        g : t :class:`torch.Tensor`
            Input tensor with values of the population parameter `g` for each feature.

        Returns
        -------
         :class:`torch.Tensor`
            The computed metric tensor, same shape as g(number of features)
        """
        return (g + 1) ** 2 / g

    @classmethod
    def model_with_sources(
        cls,
        *,
        rt: TensorOrWeightedTensor[float],
        space_shifts: TensorOrWeightedTensor[float],
        metric: TensorOrWeightedTensor[float],
        v0: TensorOrWeightedTensor[float],
        g: TensorOrWeightedTensor[float],
    ) -> torch.Tensor:
        """
        Return the model output when sources(spatial components) are present.

        Parameters
        ----------
        rt : :class:`~leaspy.uitls.weighted_tensor._weighted_tensor.TensorOrWeightedTensor`[:obj:`float`]
            Tensor containing the reparametrized time.
        space_shifts : `~leaspy.uitls.weighted_tensor._weighted_tensor.TensorOrWeightedTensor`[:obj:`float`]
            Tensor containing the values of the space-shifts
        metric :`~leaspy.uitls.weighted_tensor._weighted_tensor.TensorOrWeightedTensor`[:obj:`float`]
            Tensor containing the metric tensor used for computing the spatial/temporal influence.
        v0 : `~leaspy.uitls.weighted_tensor._weighted_tensor.TensorOrWeightedTensor`[:obj:`float`]
            Tensor containing the values of the population parameter `v0` for each feature.
        g : `~leaspy.uitls.weighted_tensor._weighted_tensor.TensorOrWeightedTensor`[:obj:`float`]
            Tensor containing the values of the population parameter `g` for each feature.

        Returns
        -------
         :class:`torch.Tensor`
            Weighted value tensor after applying sigmoid transformation,
            representing the model output with sources.
        """
        # Shape: (Ni, Nt, Nfts)
        pop_s = (None, None, ...)
        rt = unsqueeze_right(rt, ndim=1)  # .filled(float('nan'))
        w_model_logit = metric[pop_s] * (
            v0[pop_s] * rt + space_shifts[:, None, ...]
        ) - torch.log(g[pop_s])
        model_logit, weights = WeightedTensor.get_filled_value_and_weight(
            w_model_logit, fill_value=0.0
        )
        return WeightedTensor(torch.sigmoid(model_logit), weights).weighted_value





"""This module defines the `AlgorithmType`, `AlgorithmName` and `AbstractAlgo` classes"""

import inspect
import random
import sys
import time
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from typing import Generic, Optional, Type, TypeVar, Union

import numpy as np
import torch

from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.io.data import Dataset

from .settings import AlgorithmSettings, OutputsSettings

__all__ = [
    "BaseAlgorithm",
    "IterativeAlgorithm",
    "AlgorithmType",
    "AlgorithmName",
    "get_algorithm_type",
    "get_algorithm_class",
    "algorithm_factory",
    "ReturnType",
    "ModelType",
]

ModelType = TypeVar("ModelType", bound="BaseModel")
ReturnType = TypeVar("ReturnType")


class AlgorithmType(str, Enum):
    """The type of the algorithms."""

    FIT = "fit"
    PERSONALIZE = "personalize"
    SIMULATE = "simulate"


class AlgorithmName(str, Enum):
    """The available algorithms in Leaspy."""

    FIT_MCMC_SAEM = "mcmc_saem"
    FIT_LME = "lme_fit"
    PERSONALIZE_SCIPY_MINIMIZE = "scipy_minimize"
    PERSONALIZE_MEAN_POSTERIOR = "mean_posterior"
    PERSONALIZE_MODE_POSTERIOR = "mode_posterior"
    PERSONALIZE_CONSTANT = "constant_prediction"
    PERSONALIZE_LME = "lme_personalize"
    SIMULATE = "simulate"


class BaseAlgorithm(ABC, Generic[ModelType, ReturnType]):
    """Base class containing common methods for all algorithm classes.

    Parameters
    ----------
    settings : :class:`~leaspy.algo.AlgorithmSettings`
        The specifications of the algorithm as a :class:`~leaspy.algo.AlgorithmSettings` instance.

    Attributes
    ----------
    name : :class:`~leaspy.algo.base.AlgorithmName`
        Name of the algorithm.
    family : :class:`~leaspy.algo.base.AlgorithmType`
        Family of the algorithm.
    deterministic : :obj:`bool`
        True, if and only if algorithm does not involve randomness.
        Setting a seed will have no effect on such algorithms.
    algo_parameters : :obj:`dict`
        Contains the algorithm's parameters. Those are controlled by
        the :attr:`leaspy.algo.AlgorithmSettings.parameters` class attribute.
    seed : :obj:`int`, optional
        Seed used by :mod:`numpy` and :mod:`torch`.
    """

    name: AlgorithmName = None
    family: AlgorithmType = None
    deterministic: bool = False

    def __init__(self, settings: AlgorithmSettings):
        if settings.name != self.name:
            raise LeaspyAlgoInputError(
                f"Inconsistent naming: {settings.name} != {self.name}"
            )
        self.seed = settings.seed
        # we deepcopy the settings.parameters, because those algo_parameters may be
        # modified within algorithm (e.g. `n_burn_in_iter`) and we would not want the original
        # settings parameters to be also modified (e.g. to be able to reuse them without any trouble)
        self.algo_parameters = deepcopy(settings.parameters)
        self.output_manager = None

    @abstractmethod
    def set_output_manager(self, output_settings: OutputsSettings) -> None:
        raise NotImplementedError

    @staticmethod
    def _initialize_seed(seed: Optional[int]):
        """Set :mod:`random`, :mod:`numpy` and :mod:`torch` seeds and display it (static method).

        Notes - numpy seed is needed for reproducibility for the simulation algorithm which use the scipy kernel
        density estimation function. Indeed, scipy use numpy random seed.

        Parameters
        ----------
        seed : int
            The wanted seed
        """
        if seed is not None:
            random.seed(seed)
            np.random.seed(seed)
            torch.manual_seed(seed)
            # TODO: use logger instead (level=INFO)
            print(f" ==> Setting seed to {seed}")

    def run(
        self, model: ModelType, dataset: Optional[Dataset] = None, **kwargs
    ) -> ReturnType:
        """Main method, run the algorithm.

        Parameters
        ----------
        model : :class:`~leaspy.models.BaseModel`
            The used model.

        dataset : :class:`~leaspy.io.data.Dataset`
            Contains all the subjects' observations with corresponding timepoints, in torch format to speed up computations.

        Returns
        -------
        ReturnType:
            Depends on algorithm class.

        See Also
        --------
        :class:`.AbstractFitAlgo`
        :class:`.AbstractPersonalizeAlgo`
        """
        if self.algo_parameters is None:
            raise LeaspyAlgoInputError(
                f"The `{self.name}` algorithm was not properly created."
            )
        self._initialize_seed(self.seed)
        time_beginning = time.time()

        run_params = inspect.signature(self._run).parameters
        run_kwargs = {}
        if "dataset" in run_params:
            run_kwargs["dataset"] = dataset
        if "kwargs" in run_params or any(
            p.kind == inspect.Parameter.VAR_KEYWORD for p in run_params.values()
        ):
            run_kwargs.update(kwargs)

        # Appeler _run avec uniquement les arguments attendus
        output = self._run(model, **run_kwargs)
        duration_in_seconds = time.time() - time_beginning
        if self.algo_parameters.get("progress_bar"):
            print()
        print(
            f"\n{self.family.value.title()} with `{self.name}` took: {self._duration_to_str(duration_in_seconds)}"
        )
        return output

    @abstractmethod
    def _run(
        self,
        model: ModelType,
        dataset: Optional[Dataset] = None,
        **kwargs,
    ) -> ReturnType:
        """Run the algorithm (actual implementation), to be implemented in children classes.

        Parameters
        ----------
        model : :class:`~leaspy.models.BaseModel`
            The used model.

        dataset : :class:`~leaspy.io.data.Dataset`
            Contains all the subjects' observations with corresponding timepoints, in torch format to speed up computations.

        Returns
        -------
        ReturnType :
            Depends on the algorithm.

        See Also
        --------
        :class:`.AbstractFitAlgo`
        :class:`.AbstractPersonalizeAlgo`
        """
        raise NotImplementedError

    def load_parameters(self, parameters: dict):
        """Update the algorithm's parameters by the ones in the given dictionary.

        The keys in the input which does not belong to the algorithm's parameters are ignored.

        Parameters
        ----------
        parameters : :obj:`dict`
            Contains the pairs (key, value) of the requested parameters

        Examples
        --------
        >>> from leaspy.algo import AlgorithmSettings, algorithm_factory, OutputsSettings
        >>> my_algo = algorithm_factory(AlgorithmSettings("mcmc_saem"))
        >>> my_algo.algo_parameters
        {'progress_bar': True,
        'n_iter': 10000,
        'n_burn_in_iter': 9000,
        'n_burn_in_iter_frac': 0.9,
        'burn_in_step_power': 0.8,
        'random_order_variables': True,
        'sampler_ind': 'Gibbs',
        'sampler_ind_params': {'acceptation_history_length': 25,
        'mean_acceptation_rate_target_bounds': [0.2, 0.4],
        'adaptive_std_factor': 0.1},
        'sampler_pop': 'Gibbs',
        'sampler_pop_params': {'random_order_dimension': True,
        'acceptation_history_length': 25,
        'mean_acceptation_rate_target_bounds': [0.2, 0.4],
        'adaptive_std_factor': 0.1},
        'annealing': {'do_annealing': False,
         'initial_temperature': 10,
         'n_plateau': 10,
         'n_iter': None,
         'n_iter_frac': 0.5}}
        >>> parameters = {'n_iter': 5000, 'n_burn_in_iter': 4000}
        >>> my_algo.load_parameters(parameters)
        >>> my_algo.algo_parameters
        {'progress_bar': True,
        'n_iter': 5000,
        'n_burn_in_iter': 4000,
        'n_burn_in_iter_frac': 0.9,
        'burn_in_step_power': 0.8,
        'random_order_variables': True,
        'sampler_ind': 'Gibbs',
        'sampler_ind_params': {'acceptation_history_length': 25,
        'mean_acceptation_rate_target_bounds': [0.2, 0.4],
        'adaptive_std_factor': 0.1},
        'sampler_pop': 'Gibbs',
        'sampler_pop_params': {'random_order_dimension': True,
        'acceptation_history_length': 25,
        'mean_acceptation_rate_target_bounds': [0.2, 0.4],
        'adaptive_std_factor': 0.1},
        'annealing': {'do_annealing': False,
         'initial_temperature': 10,
         'n_plateau': 10,
         'n_iter': None,
         'n_iter_frac': 0.5}}
        """
        for k, v in parameters.items():
            if k in self.algo_parameters.keys():
                previous_v = self.algo_parameters[k]
                # TODO? log it instead (level=INFO or DEBUG)
                print(f"Replacing {k} parameter from value {previous_v} to value {v}")
            self.algo_parameters[k] = v

    @staticmethod
    def _duration_to_str(seconds: float, *, seconds_fmt=".0f") -> str:
        """
        Convert a float representing computation time in seconds to a string giving time in hour, minutes and
        seconds ``%h %min %s``.

        If less than one hour, do not return hours. If less than a minute, do not return minutes.

        Parameters
        ----------
        seconds : :obj:`float`
            Computation time

        Returns
        -------
        str
            Time formatting in hour, minutes and seconds.
        """
        h = int(seconds // 3600)
        m = int((seconds % 3600) // 60)
        s = seconds % 60  # float

        res = ""
        if m:
            if h:
                res += f"{h}h "
            res += f"{m}m "
        res += f"{s:{seconds_fmt}}s"

        return res

    def __str__(self) -> str:
        out = "=== ALGO ===\n"
        out += f"Instance of {self.name} algo"
        if hasattr(self, "algorithm_device"):
            out += f" [{self.algorithm_device.upper()}]"
        return out


class IterativeAlgorithm(BaseAlgorithm[ModelType, ReturnType]):
    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.current_iteration: int = 0

    @staticmethod
    def _display_progress_bar(
        iteration: int, n_iter: int, suffix: str, n_step_default: int = 50
    ):
        """
        Display a progression bar while running algorithm, simply based on `sys.stdout`.

        TODO: use tqdm instead?

        Parameters
        ----------
        iteration : :obj:`int` >= 0 or -1
            Current iteration of the algorithm.
            If a positive integer, it is the current iteration of the algorithm. If equals to '-1' then it initialises the bar.
            The final current iteration should be `n_iter - 1`
        n_iter : :obj:`int`
            Total iterations' number of the algorithm.
        suffix : :obj:`str`
            Used to differentiate types of algorithms:
                * for fit algorithms: ``suffix = 'iterations'``
                * for personalization algorithms: ``suffix = 'subjects'``.
        n_step_default : :obj:`int`, default 50
            The size of the progression bar.
        """
        n_step = min(n_step_default, n_iter)
        if iteration == -1:
            sys.stdout.write("\r")
            sys.stdout.write("|" + "-" * n_step + "|   0/%d " % n_iter + suffix)
            sys.stdout.flush()
        else:
            print_every_iter = n_iter // n_step
            iteration_plus_1 = iteration + 1
            display = iteration_plus_1 % print_every_iter
            if display == 0:
                nbar = iteration_plus_1 // print_every_iter
                sys.stdout.write("\r")
                sys.stdout.write(
                    f"|{'#' * nbar}{'-' * (n_step - nbar)}|   {iteration_plus_1}/{n_iter} {suffix}"
                )
                sys.stdout.flush()

    def _get_progress_str(self) -> Optional[str]:
        # TODO in a special mixin for sequential algos with nb of iters (MCMC fit, MCMC personalize)
        if not hasattr(self, "current_iteration"):
            return None
        return f"Iteration {self.current_iteration} / {self.algo_parameters['n_iter']}"

    def __str__(self) -> str:
        out = super().__str__()
        progress_str = self._get_progress_str()
        if progress_str:
            out += "\n" + progress_str
        return out


def get_algorithm_type(name: Union[str, AlgorithmName]) -> AlgorithmType:
    """Return the algorithm type.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`
        The name of the algorithm.

    Returns
    -------
    algorithm type: :class:`leaspy.algo.AlgorithmType`
    """
    name = AlgorithmName(name)
    if name in (AlgorithmName.FIT_LME, AlgorithmName.FIT_MCMC_SAEM):
        return AlgorithmType.FIT
    if name == AlgorithmName.SIMULATE:
        return AlgorithmType.SIMULATE
    if name in (
        AlgorithmName.PERSONALIZE_SCIPY_MINIMIZE,
        AlgorithmName.PERSONALIZE_MEAN_POSTERIOR,
        AlgorithmName.PERSONALIZE_MODE_POSTERIOR,
        AlgorithmName.PERSONALIZE_CONSTANT,
        AlgorithmName.PERSONALIZE_LME,
    ):
        return AlgorithmType.PERSONALIZE


def get_algorithm_class(name: Union[str, AlgorithmName]) -> Type[BaseAlgorithm]:
    """Return the algorithm class.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`
         The name of the algorithm.

    Returns
    -------
    algorithm class: :class:`~leaspy.algo.BaseAlgorithm`
    """
    name = AlgorithmName(name)
    if name == AlgorithmName.FIT_MCMC_SAEM:
        from .fit import TensorMcmcSaemAlgorithm

        return TensorMcmcSaemAlgorithm
    if name == AlgorithmName.FIT_LME:
        from .fit import LMEFitAlgorithm

        return LMEFitAlgorithm
    if name == AlgorithmName.PERSONALIZE_SCIPY_MINIMIZE:
        from .personalize import ScipyMinimizeAlgorithm

        return ScipyMinimizeAlgorithm
    if name == AlgorithmName.PERSONALIZE_MEAN_POSTERIOR:
        from .personalize import MeanPosteriorAlgorithm

        return MeanPosteriorAlgorithm
    if name == AlgorithmName.PERSONALIZE_MODE_POSTERIOR:
        from .personalize import ModePosteriorAlgorithm

        return ModePosteriorAlgorithm
    if name == AlgorithmName.PERSONALIZE_CONSTANT:
        from .personalize import ConstantPredictionAlgorithm

        return ConstantPredictionAlgorithm
    if name == AlgorithmName.PERSONALIZE_LME:
        from .personalize import LMEPersonalizeAlgorithm

        return LMEPersonalizeAlgorithm
    if name == AlgorithmName.SIMULATE:
        from .simulate import SimulationAlgorithm

        return SimulationAlgorithm


def algorithm_factory(settings: AlgorithmSettings) -> BaseAlgorithm:
    """Return the requested algorithm based on the provided settings.

    Parameters
    ----------
    settings : :class:`leaspy.algo.AlgorithmSettingss`
        The algorithm settings.

    Returns
    -------
    algorithm : child class of :class:`~leaspy.algo.BaseAlgorithm`
        The requested algorithm. If it exists, it will be compatible with algorithm family.
    """
    algorithm = get_algorithm_class(settings.name)(settings)
    algorithm.set_output_manager(settings.logs)
    return algorithm



from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Tuple,
    TypeVar,
)
from typing import (
    Mapping as TMapping,
)

from leaspy.utils.typing import KwargsType

__all__ = [
    "NamedInputFunction",
]

RT = TypeVar("RT")
S = TypeVar("S")


@dataclass(frozen=True)
class NamedInputFunction(Generic[RT]):
    """
    Bridge from a function with positional parameters to a function with keyword-only parameters.

    Attributes
    ----------
    f : :obj:`Callable`
        The original function.
        The named parameters to be sent in `f` should be: positional, positional-or-keyword, or variadic arguments.
        It can also have some keyword-only arguments, but they should be fixed once for all with attribute `kws`.
    parameters : :obj:`tuple` [:obj:`str`, ...]
        Assigned names, in order, for positional parameters of `f`.
    kws : None (default) or :class:`~leaspy.utils.typing.KwargsType`
        Some optional fixed keyword parameters to pass upon function calls.

    Notes
    -----
    We do not implement the mapping of keyword-only to renamed keyword-only parameters for now
    since it is not needed and would make the logic more complex. Particularly due to the existence
    of positional-only parameters.
    """

    f: Callable[..., RT]
    parameters: Tuple[str, ...]
    kws: Optional[KwargsType] = None

    def call(self, named_params: TMapping[str, Any]) -> RT:
        """
        Call the underlying function with the correct positional arguments,
        retrieved by parameter names in input variables.

        Parameters
        ----------
        named_params : :obj:`TMapping` [:obj:`str`, :obj:`Any`]
            A mapping of parameter names to their values.

        Returns
        -------
        :obj:`RT`
            The result of calling the function `f` with the provided named parameters.

        """
        # we do not enforce consistency checks on `named_params` for optimization
        # this form is especially useful when provided mapping is "lazy" / "jit-computed" (like `State`)
        return self.f(*(named_params[p] for p in self.parameters), **(self.kws or {}))

    def __call__(self, **named_params) -> RT:
        """Same as `.call()` but with variadic input.

        Returns
        -------
        :obj:`RT`
            The result of calling the function `f` with the provided named parameters.
        """
        return self.call(named_params)

    def then(self, g: Callable[[RT], S], **g_kws):  # -> NamedInputFunction[S]:
        """Return a new NamedInputFunction applying (g o f) function.

        Parameters
        ----------
        g : :obj:`Callable` [:obj:`RT`, :obj:`S`]
            A function to apply after the original function `f`.
        """

        def g_o_f(*f_args, **f_kws) -> S:
            return g(self.f(*f_args, **f_kws), **g_kws)

        # nicer for representation (too heavy otherwise)
        g_o_f.__name__ = f"{g.__name__}@{self.f.__name__}"
        g_o_f.__qualname__ = g_o_f.__name__

        return NamedInputFunction(
            f=g_o_f,
            parameters=self.parameters,
            kws=self.kws,
        )

    @staticmethod
    def bound_to(
        f: Callable[..., RT],
        check_arguments: Optional[Callable[[Tuple[str, ...], KwargsType], None]] = None,
    ):
        """Return a new factory to create new `NamedInputFunction` instances that are bound to the provided function.

        Parameters
        ----------
        f : :obj:`Callable` [:obj:`RT`]
            The function to which the new `NamedInputFunction` will be bound.

        check_arguments : :obj:`Callable` [:obj:`tuple` [:obj:`str`, ...], :obj:`~leaspy.utils.typing.KwargsType`], optional
            An optional function to check the provided parameters and keyword arguments before creating the `NamedInputFunction`.

        """

        def factory(*parameters: str, **kws) -> NamedInputFunction[RT]:
            """
            Factory of a `NamedInputFunction`, bounded to the provided function.

            Parameters
            ----------
            *parameters : :obj:`str`
                Names for positional parameters of the provided function.
            **kws
                Optional keyword-only arguments to pass to the provided function.
            """
            if check_arguments is not None:
                try:
                    check_arguments(parameters, kws)
                except Exception as e:
                    raise type(e)(f"{f.__name__}: {e}") from e
            return NamedInputFunction(f=f, parameters=parameters, kws=kws or None)

        # Nicer runtime name and docstring for the generated factory function
        factory.__name__ = f"symbolic_{f.__name__}_factory"
        factory.__qualname__ = ".".join(
            factory.__qualname__.split(".")[:-1] + [factory.__name__]
        )
        factory.__doc__ = factory.__doc__.replace(
            "the provided function", f"`{f.__name__}`"
        )

        return factory
